#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021-2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Send mail/SMS notifications with Fire data to listed subscribers.

"""

import logging
import argparse
import sys
from activefires_pp.logger import setup_logging
from activefires_pp.fire_notifications import EndUserNotifier
from activefires_pp.fire_notifications import EndUserNotifierRegional

LOG = logging.getLogger('end_user_notifier_process')

#TEST_MESSAGE = """pytroll://VIIRS/L2/MSB/nrk/utv/polar/direct_readout dataset a000680@c21856.ad.smhi.se 2020-04-23T11:46:04.299244 v1.01 application/json {"start_time": "2020-04-23T11:02:01", "end_time": "2020-04-29T11:03:25", "orbit_number": 1, "platform_name": "NOAA-20", "sensor": "viirs", "format": "EDR", "type": "NETCDF", "data_processing_level": "2", "variant": "DR", "orig_orbit_number": 12586, "dataset": [{"uri": "file:///home/a000680/Satsa/Skogsbrander/AFIMG_j01_d20200423_t1102001_e1103246_b12586_c20200423112151310071_cspp_dev.geojson", "uid": "AFIMG_j01_d20200423_t1102001_e1103246_b12586_c20200423112151310071_cspp_dev.geojson"}]}"""
TEST_MESSAGE = """pytroll://VIIRS/L2/MSB/nrk/utv/polar/direct_readout dataset a000680@c21856.ad.smhi.se 2020-04-23T11:46:04.299244 v1.01 application/json {"start_time": "2020-04-23T11:02:01", "end_time": "2020-04-29T11:03:25", "orbit_number": 1, "platform_name": "NOAA-20", "sensor": "viirs", "format": "EDR", "type": "NETCDF", "data_processing_level": "2", "variant": "DR", "orig_orbit_number": 12586, "dataset": [{"uri": "file:///home/a000680/Satsa/Skogsbrander/AFIMG_j01_d20200423_t1102001_e1103246_b12586_c20200423112151310071_cspp_dev.geojson", "uid": "AFIMG_j01_d20200423_t1102001_e1103246_b12586_c20200423112151310071_cspp_dev.geojson"}]}"""


def main():
    """Start and run the dispatcher."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--log-config",
                        help="Log config file to use instead of the standard logging.")
    parser.add_argument("-c", "--config",
                        help="YAML config file to use.")
    parser.add_argument("-n", "--netrc",
                        help="Path to .netrc file to use.")
    parser.add_argument('-r', "--regional", action='store_true',
                        help="Regional notifier - default=False")
    parser.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0,
                        help="Verbosity (between 1 and 2 occurrences with more leading to more "
                        "verbose logging). WARN=0, INFO=1, "
                        "DEBUG=2. This is overridden by the log config file if specified.")

    cmd_args = parser.parse_args()
    setup_logging(cmd_args)

    configfile = cmd_args.config
    netrcfile = cmd_args.netrc
    LOG.info("Starting up.")
    try:
        if cmd_args.regional:
            ffnotify = EndUserNotifierRegional(configfile, netrcfile=netrcfile)
        else:
            ffnotify = EndUserNotifier(configfile, netrcfile=netrcfile)

    except Exception as err:
        LOG.error('End User Notifier crashed: %s', str(err))
        sys.exit(1)
    try:
        ffnotify.start()
        ffnotify.join()
    except KeyboardInterrupt:
        LOG.debug("Interrupting")
    finally:
        ffnotify.close()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021 - 2022 Adam.Dybbro

# Author(s):

#   Adam Dybbroe <Firstname.Lastname@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Post processing on the Active Fire detections."""

import socket
from trollsift import Parser, globify
import time
import pandas as pd
from datetime import datetime, timedelta
import numpy as np
import os
from six.moves.urllib.parse import urlparse
from geojson import Feature, Point, FeatureCollection, dump
import logging
import signal
from queue import Empty
from threading import Thread
from posttroll.listener import ListenerContainer
from posttroll.message import Message
from posttroll.publisher import NoisyPublisher
import pyproj
from matplotlib.path import Path
import shapely

from activefires_pp.utils import datetime_utc2local
from activefires_pp.utils import get_local_timezone_offset
from activefires_pp.utils import json_serial
from activefires_pp.config import read_config
from activefires_pp.geometries_from_shapefiles import ShapeGeometry

# M-band output:
# column 1: latitude of fire pixel (degrees)
# column 2: longitude of fire pixel (degrees)
# column 3: M13 brightness temperature of fire pixel (K)
# column 4: Along-scan fire pixel resolution (km)
# column 5: Along-track fire pixel resolution (km)
# column 6: detection confidence (%)
#           7=low, 8=nominal, 9=high (Imager Resolution)
#           0-100 % (Moderate Resolution)
# column 7: fire radiative power (MW)
# I-band output:
# column 1: latitude of fire pixel (degrees)
# column 2: longitude of fire pixel (degrees)
# column 3: I04 brightness temperature of fire pixel (K)
# column 4: Along-scan fire pixel resolution (km)
# column 5: Along-track fire pixel resolution (km)
# column 6: detection confidence ([7,8,9]->[lo,med,hi])
# column 7: fire radiative power (MW)
#
COL_NAMES = ["latitude", "longitude", "tb", "along_scan_res", "along_track_res", "conf", "power"]


logger = logging.getLogger(__name__)
logging.getLogger("fiona").setLevel(logging.WARNING)


class ActiveFiresShapefileFiltering(object):
    """Reading, filtering and writing Active Fire detections.

    Reading either the CSPP VIIRS AF output (txt files) or the Geojson formatted files.
    Filtering for static and false alarms, and/or simply on geographical regions.
    Data is stored in geojson format.
    """

    def __init__(self, filepath=None, afdata=None, platform_name=None, timezone='GMT'):
        """Initialize the ActiveFiresShapefileFiltering class."""
        self.input_filepath = filepath
        self._afdata = afdata
        if afdata is None:
            self.metadata = {}
        else:
            self.metadata = afdata.attrs

        self.timezone = timezone
        self.platform_name = platform_name

    def get_af_data(self, filepattern=None, localtime=True):
        """Read the Active Fire results from file - ascii formatted output from CSPP VIIRS-AF."""
        if self._afdata is not None:
            # Make sure the attrs are populated with metadata instance attribute
            self._afdata.attrs.update(self.metadata)
            return self._afdata

        if not self.input_filepath or not os.path.exists(self.input_filepath):
            # FIXME! Better to raise an exception!?
            return self._afdata

        if not filepattern:
            raise AttributeError("file pattern must be provided in order to be able to read from file!")

        self.metadata = self._get_metadata_from_filename(filepattern)
        self._afdata = _read_data(self.input_filepath)
        self._add_start_and_end_time_to_active_fires_data(localtime)

        return self._afdata

    def _get_metadata_from_filename(self, infile_pattern):
        """From the filename retrieve the metadata such as satellite and sensing time."""
        return get_metadata_from_filename(infile_pattern, self.input_filepath)

    def _add_start_and_end_time_to_active_fires_data(self, localtime):
        """Add start and end time to active fires data."""
        if localtime:
            logger.info("Convert to local time zone!")
            starttime = datetime_utc2local(self.metadata['start_time'], self.timezone)
            endtime = datetime_utc2local(self.metadata['end_time'], self.timezone)
        else:
            starttime = datetime_utc2local(self.metadata['start_time'], 'GMT')
            endtime = datetime_utc2local(self.metadata['end_time'], 'GMT')

        starttime = starttime.replace(tzinfo=None)
        endtime = endtime.replace(tzinfo=None)

        self._afdata['starttime'] = np.repeat(starttime, len(self._afdata)).astype(np.datetime64)
        self._afdata['endtime'] = np.repeat(endtime, len(self._afdata)).astype(np.datetime64)

        logger.info('Start and end times: %s %s',
                    str(self._afdata['starttime'][0]),
                    str(self._afdata['endtime'][0]))

    def _apply_timezone_offset(self, obstime):
        """Apply the time zone offset to the datetime objects."""
        obstime_offset = get_local_timezone_offset(self.timezone)
        return np.repeat(obstime.replace(tzinfo=None) + obstime_offset,
                         len(self._afdata)).astype(np.datetime64)

    def fires_filtering(self, shapefile, start_geometries_index=1, inside=True):
        """Remove fires outside National borders or filter out potential false detections.

        If *inside* is True the filtering will keep those detections that are inside the polygon.
        If *inside* is False the filtering will disregard the detections that are inside the polygon.
        """
        detections = self._afdata

        lons = detections.longitude.values
        lats = detections.latitude.values

        toc = time.time()
        insides = get_global_mask_from_shapefile(shapefile, (lons, lats), start_geometries_index)
        logger.debug("Time used checking inside polygon - mpl path method: %f", time.time() - toc)

        self._afdata = detections[insides == inside]

        if len(self._afdata) == 0:
            logger.debug("No fires after filtering on Polygon...")
        else:
            logger.debug("Number of detections after filtering on Polygon: %d", len(self._afdata))

    def get_regional_filtermasks(self, shapefile, globstr):
        """Get the regional filter masks from the shapefile."""
        detections = self._afdata

        lons = detections.longitude.values
        lats = detections.latitude.values

        logger.debug("Before ShapeGeometry instance - shapefile name = %s" % str(shapefile))
        logger.debug("Shape file glob-string = %s" % str(globstr))
        shape_geom = ShapeGeometry(shapefile, globstr)
        shape_geom.load()

        p__ = pyproj.Proj(shape_geom.proj4str)
        metersx, metersy = p__(lons, lats)
        points = np.vstack([metersx, metersy]).T

        regional_masks = {}

        for attr, geometry in zip(shape_geom.attributes, shape_geom.geometries):
            test_omr = attr['Testomr']
            all_inside_test_omr = False
            some_inside_test_omr = False
            logger.debug(u'Test area: {}'.format(str(test_omr)))

            regional_masks[test_omr] = {'mask': None, 'attributes': attr}

            if isinstance(geometry, shapely.geometry.multipolygon.MultiPolygon):
                regional_masks[test_omr]['mask'] = get_mask_from_multipolygon(points, geometry)
            else:
                shape = geometry
                pth = Path(shape.exterior.coords)
                regional_masks[test_omr]['mask'] = pth.contains_points(points)

            if sum(regional_masks[test_omr]['mask']) == len(points):
                all_inside_test_omr = True
                some_inside_test_omr = True
                logger.debug("All points inside test area!")
            elif sum(regional_masks[test_omr]['mask']) > 0:
                some_inside_test_omr = True
                logger.debug("Some points inside test area!")

            regional_masks[test_omr]['all_inside_test_area'] = all_inside_test_omr
            regional_masks[test_omr]['some_inside_test_area'] = some_inside_test_omr

        return regional_masks


def _read_data(filepath):
    """Read the AF data."""
    with open(filepath, 'r') as fpt:
        return pd.read_csv(fpt, index_col=None, header=None, comment='#', names=COL_NAMES)


def get_metadata_from_filename(infile_pattern, filepath):
    """From the filename and its pattern get basic metadata of the satellite observations."""
    p__ = Parser(infile_pattern)
    fname = os.path.basename(filepath)
    try:
        res = p__.parse(fname)
    except ValueError:
        # Do something!
        return None

    # Fix the end time:
    endtime = datetime(res['start_time'].year, res['start_time'].month,
                       res['start_time'].day, res['end_hour'].hour, res['end_hour'].minute,
                       res['end_hour'].second)
    if endtime < res['start_time']:
        endtime = endtime + timedelta(days=1)

    res['end_time'] = endtime

    return res


def store(output_filename, detections):
    """Store the filtered AF detections on disk."""
    if len(detections) > 0:
        detections.to_csv(output_filename, index=False)
        return output_filename
    else:
        logger.debug("No detections to save!")
        return None


def store_geojson(output_filename, detections, platform_name=None):
    """Store the filtered AF detections in Geojson format on disk."""
    if len(detections) == 0:
        logger.debug("No detections to save!")
        return None

    # Convert points to GeoJSON
    features = []
    for idx in range(len(detections)):
        starttime = detections.iloc[idx].starttime
        endtime = detections.iloc[idx].endtime
        mean_granule_time = starttime.to_pydatetime() + (endtime.to_pydatetime() -
                                                         starttime.to_pydatetime()) / 2.

        prop = {'power': detections.iloc[idx].power,
                'tb': detections.iloc[idx].tb,
                'confidence': int(detections.iloc[idx].conf),
                'observation_time': json_serial(mean_granule_time)
                }
        if platform_name:
            prop['platform_name'] = platform_name
        else:
            logger.debug("No platform name specified for output")

        feat = Feature(
            geometry=Point(map(float, [detections.iloc[idx].longitude, detections.iloc[idx].latitude])),
            properties=prop)
        features.append(feat)

    feature_collection = FeatureCollection(features)
    path = os.path.dirname(output_filename)
    if not os.path.exists(path):
        logger.info("Create directory: %s", path)
        os.makedirs(path)

    with open(output_filename, 'w') as fpt:
        dump(feature_collection, fpt)

    return output_filename


def get_mask_from_multipolygon(points, geometry):
    """Get mask for points from a shapely Multipolygon."""
    shape = geometry.geoms[0]
    pth = Path(shape.exterior.coords)
    mask = pth.contains_points(points)

    if sum(mask) == len(points):
        return mask

    for shape in geometry.geoms[1:]:
        pth = Path(shape.exterior.coords)
        mask = np.logical_or(mask, pth.contains_points(points))
        if sum(mask) == len(points):
            break

    return mask


def get_global_mask_from_shapefile(shapefile, lonlats, start_geom_index=0):
    """Given geographical (lon,lat) points get a mask to apply when filtering."""
    lons, lats = lonlats

    logger.debug("Getting the global mask from file: shapefile file path = %s" % str(shapefile))
    shape_geom = ShapeGeometry(shapefile)
    shape_geom.load()

    p__ = pyproj.Proj(shape_geom.proj4str)

    # There is only one geometry/multi-polygon!
    geometry = shape_geom.geometries[0]

    metersx, metersy = p__(lons, lats)
    points = np.vstack([metersx, metersy]).T

    shape = geometry.geoms[0]
    pth = Path(shape.exterior.coords)
    mask = pth.contains_points(points)
    for shape in geometry.geoms[start_geom_index:]:
        pth = Path(shape.exterior.coords)
        mask = np.logical_or(mask, pth.contains_points(points))

    return mask


class ActiveFiresPostprocessing(Thread):
    """The active fires post processor."""

    def __init__(self, configfile, shp_borders, shp_mask, regional_filtermask=None):
        """Initialize the active fires post processor class."""
        super().__init__()
        self.shp_borders = shp_borders
        self.shp_filtermask = shp_mask

        self.regional_filtermask = regional_filtermask
        self.configfile = configfile
        self.options = {}

        config = read_config(self.configfile)
        self._set_options_from_config(config)

        self.host = socket.gethostname()

        self.timezone = self.options.get('timezone', 'GMT')

        self.input_topic = self.options['subscribe_topics'][0]
        self.output_topic = self.options['publish_topic']
        self.infile_pattern = self.options.get('af_pattern_ibands')
        self.outfile_pattern_national = self.options.get('geojson_file_pattern_national')
        self.outfile_pattern_regional = self.options.get('geojson_file_pattern_regional')
        self.output_dir = self.options.get('output_dir', '/tmp')

        frmt = self.options['regional_shapefiles_format']
        self.regional_shapefiles_globstr = globify(frmt)

        self.listener = None
        self.publisher = None
        self.loop = False
        self._setup_and_start_communication()

    def _setup_and_start_communication(self):
        """Set up the Posttroll communication and start the publisher."""
        logger.debug("Starting up... Input topic: %s", self.input_topic)
        now = datetime_utc2local(datetime.now(), self.timezone)
        logger.debug("Output times for timezone: {zone} Now = {time}".format(zone=str(self.timezone), time=now))

        self._check_borders_shapes_exists()

        self.listener = ListenerContainer(topics=[self.input_topic])
        self.publisher = NoisyPublisher("active_fires_postprocessing")
        self.publisher.start()
        self.loop = True
        signal.signal(signal.SIGTERM, self.signal_shutdown)

    def _set_options_from_config(self, config):
        """From the configuration on disk set the option dictionary, holding all metadata for processing."""
        for item in config:
            if not isinstance(config[item], dict):
                self.options[item] = config[item]

        if isinstance(self.options.get('subscribe_topics'), str):
            subscribe_topics = self.options.get('subscribe_topics').split(',')
            for item in subscribe_topics:
                if len(item) == 0:
                    subscribe_topics.remove(item)
            self.options['subscribe_topics'] = subscribe_topics

        if isinstance(self.options.get('publish_topics'), str):
            publish_topics = self.options.get('publish_topics').split(',')
            for item in publish_topics:
                if len(item) == 0:
                    publish_topics.remove(item)
            self.options['publish_topics'] = publish_topics

    def signal_shutdown(self, *args, **kwargs):
        """Shutdown the Active Fires postprocessing."""
        self.close()

    def run(self):
        """Run the AF post processing."""
        while self.loop:
            try:
                msg = self.listener.output_queue.get(timeout=1)
                logger.debug("Message: %s", str(msg.data))
            except Empty:
                continue
            else:
                if msg.type not in ['file', 'collection', 'dataset']:
                    logger.debug("Message type not supported: %s", str(msg.type))
                    continue

                platform_name = msg.data.get('platform_name')
                filename = get_filename_from_uri(msg.data.get('uri'))
                if not os.path.exists(filename):
                    logger.warning("File does not exist!")
                    continue

                file_ok = check_file_type_okay(msg.data.get('type'))
                no_fires_text = 'No fire detections for this granule'
                output_messages = self._generate_no_fires_messages(msg, no_fires_text)
                if not file_ok:
                    for output_msg in output_messages:
                        logger.debug("Sending message: %s", str(output_msg))
                        self.publisher.send(str(output_msg))
                    continue

                af_shapeff = ActiveFiresShapefileFiltering(filename, platform_name=platform_name,
                                                           timezone=self.timezone)
                afdata = af_shapeff.get_af_data(self.infile_pattern)

                if len(afdata) == 0:
                    logger.debug("Sending message: %s", str(output_msg))
                    self.publisher.send(str(output_msg))
                    continue

                output_messages, afdata = self.fires_filtering(msg, af_shapeff)
                logger.debug("After fires_filtering...: Number of messages = %d", len(output_messages))

                for output_msg in output_messages:
                    if output_msg:
                        logger.debug("Sending message: %s", str(output_msg))
                        self.publisher.send(str(output_msg))

                # Do the regional filtering now:
                if not self.regional_filtermask:
                    logger.info("No regional filtering is attempted.")
                    continue

                if len(afdata) == 0:
                    logger.debug("No fires - so no regional filtering to be done!")
                    continue

                # FIXME! If afdata is empty (len=0) then it seems all data are inside all regions!
                af_shapeff = ActiveFiresShapefileFiltering(afdata=afdata, platform_name=platform_name,
                                                           timezone=self.timezone)
                regional_fmask = af_shapeff.get_regional_filtermasks(self.regional_filtermask,
                                                                     globstr=self.regional_shapefiles_globstr)
                regional_messages = self.regional_fires_filtering_and_publishing(msg, regional_fmask, af_shapeff)
                for region_msg in regional_messages:
                    logger.debug("Sending message: %s", str(region_msg))
                    self.publisher.send(str(region_msg))

    def regional_fires_filtering_and_publishing(self, msg, regional_fmask, afsff_obj):
        """From the regional-fires-filter-mask and the fire detection data send regional messages."""
        logger.debug("Perform regional masking on VIIRS AF detections and publish accordingly.")

        afdata = afsff_obj.get_af_data()
        fmda = afsff_obj.metadata

        fmda['platform'] = afsff_obj.platform_name

        pout = Parser(self.outfile_pattern_regional)

        output_messages = []
        regions_with_detections = 0
        for region_name in regional_fmask:
            if not regional_fmask[region_name]['some_inside_test_area']:
                continue

            regions_with_detections = regions_with_detections + 1
            fmda['region_name'] = regional_fmask[region_name]['attributes']['Kod_omr']

            out_filepath = os.path.join(self.output_dir, pout.compose(fmda))
            logger.debug("Output file path = %s", out_filepath)
            data_in_region = afdata[regional_fmask[region_name]['mask']]
            filepath = store_geojson(out_filepath, data_in_region, platform_name=fmda['platform'])
            if not filepath:
                logger.warning("Something wrong happended storing regional " +
                               "data to Geojson - area: {name}".format(name=str(region_name)))
                continue

            outmsg = self._generate_output_message(filepath, msg, regional_fmask[region_name])
            output_messages.append(outmsg)
            logger.info("Geojson file created! Number of fires in region = %d", len(data_in_region))

        logger.debug("Regional masking done. Number of regions with fire " +
                     "detections on this granule: %s" % str(regions_with_detections))
        return output_messages

    def fires_filtering(self, msg, af_shapeff):
        """Read Active Fire data and perform spatial filtering removing false detections.

        Do the national filtering first, and then filter out potential false
        detections by the special mask for that.

        """
        logger.debug("Read VIIRS AF detections and perform quality control and spatial filtering")

        fmda = af_shapeff.metadata
        # metdata contains time and everything but it is not being transfered to the dataframe.attrs

        pout = Parser(self.outfile_pattern_national)
        out_filepath = os.path.join(self.output_dir, pout.compose(fmda))
        logger.debug("Output file path = %s", out_filepath)

        # National filtering:
        af_shapeff.fires_filtering(self.shp_borders)

        # Metadata should be transfered here!
        afdata_ff = af_shapeff.get_af_data()

        if len(afdata_ff) > 0:
            logger.debug("Doing the fires filtering: shapefile-mask = %s", str(self.shp_filtermask))
            af_shapeff.fires_filtering(self.shp_filtermask, start_geometries_index=0, inside=False)
            afdata_ff = af_shapeff.get_af_data()
            logger.debug("After fires_filtering: Number of fire detections left: %d", len(afdata_ff))

        filepath = store_geojson(out_filepath, afdata_ff, platform_name=af_shapeff.platform_name)
        out_messages = self.get_output_messages(filepath, msg, len(afdata_ff))

        return out_messages, afdata_ff

    def get_output_messages(self, filepath, msg, number_of_data):
        """Generate the adequate output message(s) depending on if an output file was created or not."""
        if filepath:
            logger.info("geojson file created! Number of fires after filtering = %d", number_of_data)
            return [self._generate_output_message(filepath, msg)]
        else:
            logger.info("No geojson file created, number of fires after filtering = %d", number_of_data)
            return self._generate_no_fires_messages(msg,
                                                    'No true fire detections inside National borders')

    def _generate_output_message(self, filepath, input_msg, region=None):
        """Create the output message to publish."""
        output_topic = generate_posttroll_topic(self.output_topic, region)
        to_send = prepare_posttroll_message(input_msg, region)
        to_send['uri'] = ('ssh://%s/%s' % (self.host, filepath))
        to_send['uid'] = os.path.basename(filepath)
        to_send['type'] = 'GEOJSON-filtered'
        to_send['format'] = 'geojson'
        to_send['product'] = 'afimg'
        pubmsg = Message(output_topic, 'file', to_send)
        return pubmsg

    def _generate_no_fires_messages(self, input_msg, msg_string):
        """Create the output messages to publish."""
        to_send = prepare_posttroll_message(input_msg)
        to_send['info'] = msg_string
        publish_messages = []
        for ext in ['National', 'Regional']:
            topic = self.output_topic + '/' + ext
            publish_messages.append(Message(topic, 'info', to_send))

        return publish_messages

    def _check_borders_shapes_exists(self):
        """Check that the national borders shapefile exists on disk."""
        if not os.path.exists(self.shp_borders):
            raise OSError("Shape file does not exist! Filename = %s" % self.shp_borders)

    def close(self):
        """Shutdown the Active Fires postprocessing."""
        logger.info('Terminating Active Fires post processing.')
        self.loop = False
        try:
            self.listener.stop()
        except Exception:
            logger.exception("Couldn't stop listener.")
        if self.publisher:
            try:
                self.publisher.stop()
            except Exception:
                logger.exception("Couldn't stop publisher.")


def check_file_type_okay(file_type):
    """Check if the file is of the correct type."""
    if file_type not in ['txt', 'TXT']:
        logger.info('File type not txt: %s', str(file_type))
        return False
    return True


def get_filename_from_uri(uri):
    """Get the file name from the uri given."""
    logger.info('File uri: %s', str(uri))
    url = urlparse(uri)
    return url.path


def generate_posttroll_topic(output_topic, region=None):
    """Create the topic for the posttroll message to publish."""
    if region:
        output_topic = output_topic + '/Regional/' + region['attributes']['Kod_omr']
    else:
        output_topic = output_topic + '/National'

    return output_topic


def prepare_posttroll_message(input_msg, region=None):
    """Create the basic posttroll-message fields and return."""
    to_send = input_msg.data.copy()
    to_send.pop('dataset', None)
    to_send.pop('collection', None)
    to_send.pop('uri', None)
    to_send.pop('uid', None)
    to_send.pop('format', None)
    to_send.pop('type', None)
    # FIXME! Check that the region_name is stored as a unicode string!
    if region:
        to_send['region_name'] = region['attributes']['Testomr']
        to_send['region_code'] = region['attributes']['Kod_omr']

    return to_send



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, 2022 Adam.Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Loading shapefile geometries from files."""

from glob import glob
import os
import cartopy.io.shapereader as shpreader
import pycrs


class ShapeGeometry(object):
    """Geometry from a shape file."""

    def __init__(self, shapefilepath, globstr='*.shp'):
        """Initialize the ShapeGeometry class."""
        self.filepaths = _get_shapefile_paths(shapefilepath, globstr)
        self.geometries = None
        self.attributes = None
        self._get_proj()

    def load(self):
        """Load the geometries and the associated attributes."""
        records = []
        for filepath in self.filepaths:
            records = records + [r for r in shpreader.Reader(filepath).records()]

        self._records = records
        self._load_member_from_records('geometries', 'geometry')
        self._load_member_from_records('attributes', 'attributes')

    def _get_proj(self):
        """Get and return the Proj.4 string."""
        self.proj4str = []
        for filepath in self.filepaths:
            prj_filename = filepath.strip('.shp') + '.prj'
            crs = pycrs.load.from_file(prj_filename)
            if crs.name == 'SWEREF99_TM' and crs.proj.name.proj4 == 'utm':
                utm_zone_proj4 = ' +zone=33'
                proj4str = crs.to_proj4() + utm_zone_proj4
            else:
                proj4str = crs.to_proj4()
            self.proj4str.append(proj4str)

        first_proj4_str = self.proj4str[0]
        for proj4_str in self.proj4str[1:]:
            if proj4_str != first_proj4_str:
                return

        self.proj4str = first_proj4_str

    def _load_member_from_records(self, class_member, record_type):
        """Load a member of the shapely.geometry object and set the corresponding class member."""
        setattr(self, class_member, [getattr(rec, record_type) for rec in self._records])


def _get_shapefile_paths(path, globstr='*.shp'):
    """Get full filepaths for all shapefiles in directory or simply return the paths as a list.

    From a path to a directory with shapefiles or a full file path,
    return list of file paths for all shapefiles.
    """
    if os.path.isfile(path):
        return [path]

    shapefile_paths = glob(os.path.join(path, globstr))
    if len(shapefile_paths) == 0:
        raise OSError('No matching shapefiles found on disk. Path = %s, glob-string = %s', path, globstr)
    return shapefile_paths



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Post geojson formatted Alarms to a ReST-API."""

import logging
import requests

# Data payload to be posted - Example:
# {"type": "Feature", "geometry": {"type": "Point", "coordinates": [15.860621, 61.403141]},
# "properties": {"power": 3.09576535, "tb": 328.81933594, "confidence": 8,
#                "observation_time": "2022-08-02T03:27:43.850000",
#                "platform_name": "NOAA-20",  "related_detection": false}}

LOG = logging.getLogger(__name__)


def post_alarm(geojson_data, api_url, xauth=None):
    """Post an Alarm to a rest-api stored as a geojson file."""
    if xauth is None:
        headers = {"Content-Type": "application/json; charset=utf-8"}
    else:
        headers = {"Content-Type": "application/json; charset=utf-8",
                   "x-auth-satellite-alarm": xauth}

    response = requests.post(api_url,
                             headers=headers,
                             json=geojson_data)

    LOG.info("Alarm posted: Response = %s", str(response))
    LOG.debug("Status code = %d", response.status_code)
    response.raise_for_status()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021-2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import cartopy.io.shapereader as shpreader
from datetime import date, datetime, timezone
from urllib.parse import urlparse
import pathlib
import logging
import zoneinfo

LOG = logging.getLogger(__name__)


def datetime_utc2local(utc_dtime, tzone_str, is_dst=True):
    """Convert a UTC datetime to local time, using DST on default."""
    tzone = zoneinfo.ZoneInfo(tzone_str)
    utc_dtime = utc_dtime.replace(tzinfo=timezone.utc)
    return utc_dtime.astimezone(tzone)


def get_local_timezone_offset(timezone_str):
    """Get the local time zone offset as a timedelta object."""
    utcnow = datetime.utcnow()
    utcnow = utcnow.replace(tzinfo=timezone.utc)
    tzone = zoneinfo.ZoneInfo(timezone_str)
    return utcnow.astimezone(tzone).replace(tzinfo=timezone.utc) - utcnow


def get_geometry_from_shapefile(shapefile):
    """Read shapefile and return geometry as a multipolygon"""
    records = shpreader.Reader(shapefile).records()
    geometries = [c.geometry for c in records]

    return geometries


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def get_filename_from_posttroll_message(pytroll_message):
    """Get the filename from the Posttroll message."""
    url = urlparse(pytroll_message.data.get('uri'))
    filepath = pathlib.Path(url.path)
    LOG.info('File path: %s', str(filepath))
    return filepath



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Adam Dybbroe at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the fire notifications.
"""

import unittest
from unittest.mock import patch
import yaml
import io
from posttroll.message import Message

from activefires_pp.fire_notifications import EndUserNotifier
from activefires_pp.fire_notifications import EndUserNotifierRegional
from activefires_pp.fire_notifications import get_recipients_for_region

TEST_CONFIG_FILE = "/home/a000680/usr/src/forks/activefires-pp/examples/fire_notifier.yaml"
TEST_CONFIG_FILE_REGIONAL = "/home/a000680/usr/src/forks/activefires-pp/examples/fire_notifier_regional.yaml"

NAT_CONFIG = """
# Publish/subscribe
publish_topic: VIIRS/L2/MSB/National
subscribe_topics: VIIRS/L2/Fires/PP/National

smtp_server: smtp.mydomain.se

domain: mydomain.se

sender: active-fires@mydomain.se

recipients: ["recipient1@recipients.se", "recipient2@recipients.se", "recipient3@recipients.se"]
recipients_attachment: ["recipient1@recipients.se", "recipient2@recipients.se"]
subject: "My subject"

max_number_of_fires_in_sms: 3

fire_data:
  - power
  - observation_time

unsubscribe:
  address: unsubscribe@mydomain.xx
  text: 'Stop being bothered: Send a note to unsubscribe@mydomain.xx'
"""

REG_CONFIG = """# Publish/subscribe
publish_topic: VIIRS/L2/MSB/Regional
subscribe_topics: VIIRS/L2/Fires/PP/Regional

smtp_server: smtp.mydomain.xx

domain: mydomain.xx

sender: active-fires@mydomain.xx

recipients:
  Area1-name:
    subject: "My subject"
    Kod_omr: '0999'
    name: 'Name of my area 1'
    recipients:
      - active-fires-sms-0999@mydomain.xx
    recipients_attachment: [active-fires-0999@mydomain.xx, ]

  Area2-name:
    subject: "My subject"
    Kod_omr: '0114'
    name: 'Name of my area 2'
    recipients: [active-fires-sms-0998@mydomain.xx, ]
    recipients_attachment: [active-fires-0998@mydomain.xx, ]

max_number_of_fires_in_sms: 3

fire_data:
  - power
  - observation_time

unsubscribe:
  address: unsubscribe@mydomain.xx
  text: 'Stop being bothered: Send a note to unsubscribe@mydomain.xx'
"""

REGIONAL_TEST_MESSAGE = """pytroll://VIIRS/L2/Fires/PP/Regional/0114 file safusr.u@lxserv1043.smhi.se 2021-04-19T11:16:49.542021 v1.01 application/json {"start_time": "2021-04-16T12:29:53", "end_time": "2021-04-16T12:31:18", "orbit_number": 1, "platform_name": "NOAA-20", "sensor": "viirs", "data_processing_level": "2", "variant": "DR", "orig_orbit_number": 17666, "region_name": "Storstockholms brandf\u00f6rsvar", "region_code": "0114", "uri": "ssh://lxserv1043.smhi.se//san1/polar_out/direct_readout/viirs_active_fires/filtered/AFIMG_NOAA-20_d20210416_t122953_0114.geojson", "uid": "AFIMG_NOAA-20_d20210416_t122953_0114.geojson", "type": "GEOJSON-filtered", "format": "geojson", "product": "afimg"}"""

NATIONAL_TEST_MESSAGE = """pytroll://VIIRS/L2/Fires/PP/National file safusr.u@lxserv1043.smhi.se 2021-04-19T11:16:49.519087 v1.01 application/json {"start_time": "2021-04-16T12:29:53", "end_time": "2021-04-16T12:31:18", "orbit_number": 1, "platform_name": "NOAA-20", "sensor": "viirs", "data_processing_level": "2", "variant": "DR", "orig_orbit_number": 17666, "uri": "ssh://lxserv1043.smhi.se//san1/polar_out/direct_readout/viirs_active_fires/filtered/AFIMG_j01_d20210416_t122953.geojson", "uid": "AFIMG_j01_d20210416_t122953.geojson", "type": "GEOJSON-filtered", "format": "geojson", "product": "afimg"}"""


class MyNetrcMock(object):

    def __init__(self):

        self.hosts = {'default': ('my_user', None, 'my_passwd')}

    def authenticators(self, host):
        return self.hosts.get(host)


class TestNotifyEndUsers(unittest.TestCase):

    @patch('activefires_pp.fire_notifications.netrc')
    @patch('activefires_pp.fire_notifications.socket.gethostname')
    @patch('activefires_pp.fire_notifications.read_config')
    @patch('activefires_pp.fire_notifications.EndUserNotifier._setup_and_start_communication')
    def test_get_options_national_filtering(self, setup_comm, read_config, gethostname, netrc):

        secrets = MyNetrcMock()
        netrc.return_value = secrets
        gethostname.return_value = 'default'

        myconfigfile = "/my/config/file/path"
        natstream = io.StringIO(NAT_CONFIG)

        read_config.return_value = yaml.load(natstream, Loader=yaml.UnsafeLoader)

        this = EndUserNotifier(myconfigfile)

        expected = {'publish_topic': 'VIIRS/L2/MSB/National',
                    'subscribe_topics': ['VIIRS/L2/Fires/PP/National'],
                    'smtp_server': 'smtp.mydomain.se', 'domain': 'mydomain.se', 'sender': 'active-fires@mydomain.se',
                    'recipients': ['recipient1@recipients.se', 'recipient2@recipients.se', 'recipient3@recipients.se'],
                    'recipients_attachment': ['recipient1@recipients.se', 'recipient2@recipients.se'],
                    'subject': 'My subject', 'max_number_of_fires_in_sms': 3,
                    'fire_data': ['power', 'observation_time'],
                    'unsubscribe': {'address': 'unsubscribe@mydomain.xx',
                                    'text': 'Stop being bothered: Send a note to unsubscribe@mydomain.xx'},
                    'unsubscribe_address': 'unsubscribe@mydomain.xx',
                    'unsubscribe_text': 'Stop being bothered: Send a note to unsubscribe@mydomain.xx'}

        self.assertDictEqual(expected, this.options)

        assert this.smtp_server == 'smtp.mydomain.se'
        assert this.domain == 'mydomain.se'
        assert this.sender == 'active-fires@mydomain.se'

        self.assertListEqual(this.recipients.recipients_all,
                             ['recipient1@recipients.se', 'recipient2@recipients.se', 'recipient3@recipients.se'])
        self.assertListEqual(this.recipients.recipients_with_attachment,
                             ['recipient1@recipients.se', 'recipient2@recipients.se'])
        self.assertListEqual(this.recipients.recipients_without_attachment, ['recipient3@recipients.se'])

        assert this.subject == 'My subject'
        assert this.max_number_of_fires_in_sms == 3
        self.assertListEqual(this.fire_data, ['power', 'observation_time'])
        assert this.unsubscribe_address == 'unsubscribe@mydomain.xx'
        assert this.input_topic == 'VIIRS/L2/Fires/PP/National'
        assert this.output_topic == 'VIIRS/L2/MSB/National'


class TestNotifyEndUsersRegional(unittest.TestCase):

    @patch('activefires_pp.fire_notifications.netrc')
    @patch('activefires_pp.fire_notifications.socket.gethostname')
    @patch('activefires_pp.fire_notifications.read_config')
    @patch('activefires_pp.fire_notifications.EndUserNotifierRegional._setup_and_start_communication')
    def test_get_options_regional_filtering(self, setup_comm, read_config, gethostname, netrc):

        secrets = MyNetrcMock()
        netrc.return_value = secrets
        gethostname.return_value = 'default'

        myconfigfile = "/my/config/file/path"
        regstream = io.StringIO(REG_CONFIG)

        read_config.return_value = yaml.load(regstream, Loader=yaml.UnsafeLoader)

        this = EndUserNotifierRegional(myconfigfile)

        expected = {'publish_topic': 'VIIRS/L2/MSB/Regional',
                    'subscribe_topics': ['VIIRS/L2/Fires/PP/Regional'],
                    'smtp_server': 'smtp.mydomain.xx', 'domain': 'mydomain.xx',
                    'sender': 'active-fires@mydomain.xx',
                    'recipients': {
                        'Area1-name': {'subject': 'My subject', 'Kod_omr': '0999', 'name': 'Name of my area 1',
                                       'recipients': ['active-fires-sms-0999@mydomain.xx'],
                                       'recipients_attachment': ['active-fires-0999@mydomain.xx']},
                        'Area2-name': {'subject': 'My subject', 'Kod_omr': '0114', 'name':
                                       'Name of my area 2', 'recipients': ['active-fires-sms-0998@mydomain.xx'],
                                       'recipients_attachment': ['active-fires-0998@mydomain.xx']}},
                    'max_number_of_fires_in_sms': 3, 'fire_data': ['power', 'observation_time'],
                    'unsubscribe': {'address': 'unsubscribe@mydomain.xx',
                                    'text': 'Stop being bothered: Send a note to unsubscribe@mydomain.xx'},
                    'unsubscribe_address': 'unsubscribe@mydomain.xx',
                    'unsubscribe_text': 'Stop being bothered: Send a note to unsubscribe@mydomain.xx'}

        self.assertDictEqual(expected, this.options)

        assert this.smtp_server == 'smtp.mydomain.xx'
        assert this.domain == 'mydomain.xx'
        assert this.sender == 'active-fires@mydomain.xx'

        # assert len(this.recipients) == 2
        # recipients = this.recipients['Area1-name']

        # expected = {'subject': 'My subject', 'knkod': '0999', 'name': 'Name of my area 1',
        #             'recipients': ['active-fires-sms-0999@mydomain.xx'],
        #             'recipients_attachment': ['active-fires-0999@mydomain.xx']}
        # self.assertDictEqual(expected, recipients)

        # recipients = this.recipients['Area2-name']
        # expected = {'subject': 'My subject', 'knkod': '0114', 'name': 'Name of my area 2',
        #             'recipients': ['active-fires-sms-0998@mydomain.xx'],
        #             'recipients_attachment': ['active-fires-0998@mydomain.xx']}
        # self.assertDictEqual(expected, recipients)

        # assert this.subject is None
        # assert this.max_number_of_fires_in_sms == 3
        # self.assertListEqual(this.fire_data, ['power', 'observation_time'])
        # assert this.unsubscribe_address == 'unsubscribe@mydomain.xx'
        # assert this.input_topic == 'VIIRS/L2/Fires/PP/Regional'
        # assert this.output_topic == 'VIIRS/L2/MSB/Regional'

        input_msg = Message.decode(rawstr=REGIONAL_TEST_MESSAGE)
        # this.notify_end_users(input_msg)

        # ffdata = {"features": [{"geometry": {"coordinates": [17.198933, 59.577972], "type": "Point"},
        #                        "properties": {"observation_time": "2021-04-16T14:30:35.900000",
        #                                       "platform_name": "NOAA-20", "power": 5.53501701, "tb": 367.0},
        #                        "type": "Feature"}], "type": "FeatureCollection"}

    @patch('activefires_pp.fire_notifications.netrc')
    @patch('activefires_pp.fire_notifications.socket.gethostname')
    @patch('activefires_pp.fire_notifications.read_config')
    @patch('activefires_pp.fire_notifications.EndUserNotifierRegional._setup_and_start_communication')
    def test_get_recipients_for_region(self, setup_comm, read_config, gethostname, netrc):
        """Test getting the recipients for a given region."""

        secrets = MyNetrcMock()
        netrc.return_value = secrets
        gethostname.return_value = 'default'

        myconfigfile = "/my/config/file/path"
        regstream = io.StringIO(REG_CONFIG)

        read_config.return_value = yaml.load(regstream, Loader=yaml.UnsafeLoader)

        this = EndUserNotifierRegional(myconfigfile)

        recipients = this.options.get('recipients')
        region_code = '0114'
        result = get_recipients_for_region(recipients, region_code)

        assert result.region_name == 'Name of my area 2'
        assert result.region_code == '0114'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c21856.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Fixtures for unittests."""

import pytest


TEST_YAML_CONFIG_CONTENT = """# Publish/subscribe
subscribe_topics: /VIIRS/L2/Fires/PP/National
publish_topic: /VIIRS/L2/Fires/PP/SOSAlarm

geojson_file_pattern_alarms: sos_{start_time:%Y%m%d_%H%M%S}_{id:d}.geojson

fire_alarms_dir: /path/where/the/filtered/alarms/will/be/stored

restapi_url: "https://xxx.smhi.se:xxxx"

time_and_space_thresholds:
  hour_threshold: 6
  long_fires_threshold_km: 1.2
  spatial_threshold_km: 0.8
"""

TEST_POST_PROCESSING_YAML_CONFIG_CONTENT = """# Publish/subscribe
publish_topic: /VIIRS/L2/Fires/PP
subscribe_topics: VIIRS/L2/AFI

af_pattern_ibands: AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S%f}_e{end_hour:%H%M%S%f}_b{orbit:s}_c{processing_time:%Y%m%d%H%M%S%f}_cspp_dev.txt

geojson_file_pattern_national: AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S}.geojson
geojson_file_pattern_regional: AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S}_{region_name:s}.geojson

regional_shapefiles_format: omr_{region_code:s}_Buffer.{ext:s}

output_dir: /path/where/the/filtered/results/will/be/stored

timezone: Europe/Stockholm
"""  # noqa

TEST_YAML_TOKENS = """xauth_tokens:
  x-auth-satellite-alarm : 'my-token'
"""

# AFIMG_NOAA-20_20210619_005803_sweden.geojson
TEST_GEOJSON_FILE_CONTENT_MONSTERAS = """{"type": "FeatureCollection", "features":
[{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.240452, 57.17329]},
"properties": {"power": 4.19946575, "tb": 336.38024902, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.247334, 57.172443]},
"properties": {"power": 5.85325146, "tb": 339.84768677, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.242519, 57.17498]},
"properties": {"power": 3.34151864, "tb": 316.57772827, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.249384, 57.174122]},
"properties": {"power": 3.34151864, "tb": 310.37808228, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.241102, 57.171574]},
"properties": {"power": 3.34151864, "tb": 339.86465454, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.247967, 57.170712]},
"properties": {"power": 3.34151864, "tb": 335.95074463, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.246538, 57.167309]},
"properties": {"power": 3.10640526, "tb": 337.62503052, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.239674, 57.168167]},
"properties": {"power": 3.10640526, "tb": 305.36495972, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.245104, 57.163902]},
"properties": {"power": 3.10640526, "tb": 336.21279907, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.251965, 57.16304]},
"properties": {"power": 2.40693879, "tb": 306.66555786, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.250517, 57.159637]},
"properties": {"power": 2.23312426, "tb": 325.92211914, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.24366, 57.160496]},
"properties": {"power": 1.51176202, "tb": 317.16009521, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.242212, 57.157097]},
"properties": {"power": 1.51176202, "tb": 303.77804565, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.249069, 57.156235]},
"properties": {"power": 2.23312426, "tb": 310.37322998, "confidence": 8,
"observation_time": "2021-06-19T02:58:45.700000+02:00", "platform_name": "NOAA-20"}}]}"""

TEST_GEOJSON_FILE_CONTENT = """{"type": "FeatureCollection", "features":
[{"type": "Feature", "geometry": {"type": "Point", "coordinates": [23.562864, 67.341919]},
"properties": {"power": 1.62920368, "tb": 325.2354126, "confidence": 8,
"observation_time": "2022-06-29T14:01:08.850000", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [23.56245, 67.347328]},
"properties": {"power": 3.40044808, "tb": 329.46963501, "confidence": 8,
"observation_time": "2022-06-29T14:01:08.850000", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [23.555086, 67.343231]},
"properties": {"power": 6.81757641, "tb": 334.62347412, "confidence": 8,
"observation_time": "2022-06-29T14:01:08.850000", "platform_name": "NOAA-20"}}]}"""

# Past alarms:
PAST_ALARMS_MONSTERAS1 = """{"type": "FeatureCollection", "features":
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.246222, 57.175987]},
"properties": {"power": 1.83814871, "tb": 302.3949585, "confidence": 8,
"observation_time": "2021-06-19T02:07:33.050000+02:00", "platform_name": "Suomi-NPP", "related_detection": true}}}"""

PAST_ALARMS_MONSTERAS2 = """{"type": "FeatureCollection", "features":
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [16.245516, 57.1651]},
"properties": {"power": 2.94999027, "tb": 324.5098877, "confidence": 8,
"observation_time": "2021-06-19T02:07:33.050000+02:00", "platform_name": "Suomi-NPP", "related_detection": true}}}"""

PAST_ALARMS_MONSTERAS3 = """{"features": {"geometry": {"coordinates": [16.252192, 57.15242], "type": "Point"},
"properties": {"confidence": 8, "observation_time": "2021-06-18T14:49:01.750000+02:00",
"platform_name": "NOAA-20", "related_detection": false, "power": 2.87395763, "tb": 330.10293579},
"type": "Feature"}, "type": "FeatureCollection"}"""


@pytest.fixture
def fake_token_file(tmp_path):
    """Write fake token file."""
    file_path = tmp_path / '.sometokenfile.yaml'
    with open(file_path, 'w') as fpt:
        fpt.write(TEST_YAML_TOKENS)

    yield file_path


@pytest.fixture
def fake_yamlconfig_file(tmp_path):
    """Write fake yaml config file."""
    file_path = tmp_path / 'test_alarm_filtering_config.yaml'
    with open(file_path, 'w') as fpt:
        fpt.write(TEST_YAML_CONFIG_CONTENT)

    yield file_path


@pytest.fixture
def fake_yamlconfig_file_post_processing(tmp_path):
    """Write fake yaml config file."""
    file_path = tmp_path / 'test_af_post_processing_config.yaml'
    with open(file_path, 'w') as fpt:
        fpt.write(TEST_POST_PROCESSING_YAML_CONFIG_CONTENT)

    yield file_path


@pytest.fixture
def fake_geojson_file_many_detections(tmp_path):
    """Write fake geojson file with many close detections."""
    file_path = tmp_path / 'test_afimg_NOAA-20_20210619_005803_sweden.geojson'
    with open(file_path, 'w') as fpt:
        fpt.write(TEST_GEOJSON_FILE_CONTENT_MONSTERAS)

    yield file_path


@pytest.fixture
def fake_geojson_file(tmp_path):
    """Write fake geojson file."""
    file_path = tmp_path / 'test_afimg_20220629_120026.geojson'
    with open(file_path, 'w') as fpt:
        fpt.write(TEST_GEOJSON_FILE_CONTENT)

    yield file_path


@pytest.fixture
def fake_past_detections_dir(tmp_path):
    """Create fake directory with past detections."""
    past_detections_dir = tmp_path / 'past_detections'
    past_detections_dir.mkdir()
    file_path = past_detections_dir / 'sos_20210619_000651_0.geojson'
    with open(file_path, 'w') as fpt:
        fpt.write(PAST_ALARMS_MONSTERAS1)

    file_path = past_detections_dir / 'sos_20210619_000651_1.geojson'
    with open(file_path, 'w') as fpt:
        fpt.write(PAST_ALARMS_MONSTERAS2)

    file_path = past_detections_dir / 'sos_20210618_124819_0.geojson'
    with open(file_path, 'w') as fpt:
        fpt.write(PAST_ALARMS_MONSTERAS2)

    yield file_path.parent


@pytest.fixture
def fake_national_borders_shapefile(tmp_path):
    """Write fake national borders shape file."""
    file_path = tmp_path / 'some_national_borders_shape.yaml'
    with open(file_path, 'w') as fpt:
        fpt.write('')

    yield file_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam.Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import logging
import argparse
import sys
from activefires_pp.logger import setup_logging
from activefires_pp.spatiotemporal_alarm_filtering import AlarmFilterRunner

LOG = logging.getLogger('spatiotemporal_alarm_filtering')


def main():
    """Start and run the Spatio-temporal fire detection alarm filtering."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--log-config",
                        help="Log config file to use instead of the standard logging.")
    parser.add_argument("-c", "--config",
                        help="YAML config file to use.")
    parser.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0,
                        help="Verbosity (between 1 and 2 occurrences with more leading to more "
                        "verbose logging). WARN=0, INFO=1, "
                        "DEBUG=2. This is overridden by the log config file if specified.")

    cmd_args = parser.parse_args()
    setup_logging(cmd_args)

    configfile = cmd_args.config
    LOG.info("Starting up.")
    try:
        ffnotify = AlarmFilterRunner(configfile)
    except Exception as err:
        LOG.error('Alarm Filter Runner crashed: %s', str(err))
        sys.exit(1)
    try:
        ffnotify.start()
        ffnotify.join()
    except KeyboardInterrupt:
        LOG.debug("Interrupting")
    finally:
        ffnotify.close()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the Geojson utillities."""

from activefires_pp.geojson_utils import read_geojson_data
from activefires_pp.geojson_utils import get_geojson_files_in_observation_time_order
from activefires_pp.geojson_utils import store_geojson_alarm
from trollsift import Parser

from datetime import datetime
import pytest
import logging

TEST_GEOJSON_FILE_CONTENT = """{"type": "FeatureCollection", "features":
[{"type": "Feature", "geometry": {"type": "Point", "coordinates": [23.562864, 67.341919]},
"properties": {"power": 1.62920368, "tb": 325.2354126, "confidence": 8,
"observation_time": "2022-06-29T14:01:08.850000", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [23.56245, 67.347328]},
"properties": {"power": 3.40044808, "tb": 329.46963501, "confidence": 8,
"observation_time": "2022-06-29T14:01:08.850000", "platform_name": "NOAA-20"}},
{"type": "Feature", "geometry": {"type": "Point", "coordinates": [23.555086, 67.343231]},
"properties": {"power": 6.81757641, "tb": 334.62347412, "confidence": 8,
"observation_time": "2022-06-29T14:01:08.850000", "platform_name": "NOAA-20"}}]}"""


@pytest.fixture
def fake_geojson_file(tmp_path):
    """Write fake geojson file."""
    file_path = tmp_path / 'test_afimg_20220629_120026.geojson'
    with open(file_path, 'w') as fpt:
        fpt.write(TEST_GEOJSON_FILE_CONTENT)

    yield file_path


@pytest.fixture
def fake_empty_geojson_file(tmp_path):
    """Write fake empty geojson file."""
    file_path = tmp_path / 'test_afimg_20220629_110913.geojson'
    file_path.touch()
    yield file_path


@pytest.fixture
def fake_nonexisting_geojson_file(tmp_path):
    """Write non existing (geojson) file."""
    file_path = tmp_path / 'test_afimg_20220629_110913.geojson'
    yield file_path


@pytest.fixture
def fake_past_detections_dir(tmp_path):
    """Create fake directory with past detections."""
    past_detections_dir = tmp_path / 'past_detections'
    past_detections_dir.mkdir()
    file_path = past_detections_dir / 'sos_20210619_005803_0.geojson'
    file_path.touch()
    file_path = past_detections_dir / 'sos_20210619_000651_0.geojson'
    file_path.touch()
    file_path = past_detections_dir / 'sos_20210619_000651_1.geojson'
    file_path.touch()
    file_path = past_detections_dir / 'sos_20210618_124819_0.geojson'
    file_path.touch()
    file_path = past_detections_dir / 'sos_20210618_110719_0.geojson'
    file_path.touch()

    yield file_path.parent


def test_read_and_get_geojson_data_from_file(fake_geojson_file):
    """Test reading a geojson file and return the content."""
    ffdata = read_geojson_data(fake_geojson_file)

    assert len(ffdata) == 2
    assert ffdata['type'] == 'FeatureCollection'
    assert len(ffdata['features']) == 3
    feature1 = ffdata['features'][0]
    assert 'type' in feature1.keys()
    assert 'geometry' in feature1.keys()
    assert 'properties' in feature1.keys()
    assert feature1['type'] == 'Feature'
    assert feature1['geometry'] == {"coordinates": [23.562864, 67.341919], "type": "Point"}
    assert feature1['properties'] == {"confidence": 8, "observation_time": "2022-06-29T14:01:08.850000",
                                      "platform_name": "NOAA-20", "power": 1.62920368, "tb": 325.2354126}


def test_read_and_get_geojson_data_from_nonexisting_file(caplog, fake_nonexisting_geojson_file):
    """Test reading a geojson file when file is not there."""
    with caplog.at_level(logging.ERROR):
        ffdata = read_geojson_data(fake_nonexisting_geojson_file)

    assert ffdata is None
    log_output = "No valid filename to read: {filename}".format(filename=fake_nonexisting_geojson_file)
    assert log_output in caplog.text


def test_read_and_get_geojson_data_from_empty_file(caplog, fake_empty_geojson_file):
    """Test reading an empty geojson file and return the content."""
    with caplog.at_level(logging.WARNING):
        ffdata = read_geojson_data(fake_empty_geojson_file)

    assert ffdata is None
    log_output = "Geojson file invalid and cannot be read: {filename}".format(filename=fake_empty_geojson_file)
    assert log_output in caplog.text


def test_get_geojson_files_in_observation_time_order(fake_past_detections_dir):
    """Test getting the list of geojson files ordered by observation time - most recent file first."""
    starttime = datetime(2021, 6, 18, 12, 0)
    endtime = datetime(2021, 6, 19, 0, 30)
    # geojson_file_pattern_alarms: sos_{start_time:%Y%m%d_%H%M%S}_{id:d}.geojson
    pattern = "sos_{start_time:%Y%m%d_%H%M%S}_{id:d}.geojson"

    recent = get_geojson_files_in_observation_time_order(fake_past_detections_dir, pattern, (starttime, endtime))
    assert set(recent) == {'sos_20210618_124819_0.geojson', 'sos_20210619_000651_1.geojson',
                           'sos_20210619_000651_0.geojson'}

    starttime = datetime(2021, 6, 18, 12, 0)
    endtime = datetime(2021, 6, 19, 12, 0)
    recent = get_geojson_files_in_observation_time_order(fake_past_detections_dir, pattern, (starttime, endtime))

    assert set(recent) == {'sos_20210618_124819_0.geojson', 'sos_20210619_000651_1.geojson',
                           'sos_20210619_000651_0.geojson', 'sos_20210619_005803_0.geojson'}

    starttime = datetime(2022, 6, 18, 12, 0)
    endtime = datetime(2022, 6, 19, 12, 0)
    recent = get_geojson_files_in_observation_time_order(fake_past_detections_dir, pattern, (starttime, endtime))
    assert len(recent) == 0


def test_store_geojson_alarm(fake_past_detections_dir):
    """Test storing the geojson alarm object."""
    sos_alarms_file_pattern = 'sos_{start_time:%Y%m%d_%H%M%S}_{id:d}.geojson'
    file_parser = Parser(sos_alarms_file_pattern)
    idx = 0
    alarm = {"features": {"geometry": {"coordinates": [16.249069, 57.156235], "type": "Point"},
                          "properties": {"confidence": 8, "observation_time": "2021-06-19T02:58:45.700000+02:00",
                                         "platform_name": "NOAA-20",
                                         "power": 2.23312426,
                                         "related_detection": False,
                                         "tb": 310.37322998}, "type": "Feature"},
             "type": "FeatureCollection"}

    result_filename = store_geojson_alarm(fake_past_detections_dir, file_parser, idx, alarm)
    assert result_filename.exists() is True
    assert result_filename.name == 'sos_20210619_005845_0.geojson'

    json_test_data = read_geojson_data(result_filename)

    assert json_test_data['features']['geometry']['coordinates'] == [16.249069, 57.156235]
    assert json_test_data['features']['properties']['confidence'] == 8
    assert json_test_data['features']['properties']['observation_time'] == "2021-06-19T02:58:45.700000+02:00"
    assert json_test_data['features']['properties']['platform_name'] == "NOAA-20"
    assert json_test_data['features']['properties']['power'] == 2.23312426
    assert json_test_data['features']['properties']['related_detection'] is False
    assert json_test_data['features']['properties']['tb'] == 310.37322998



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Spatial and temporal filtering of fire alarms.

The aim of this filtering is to avoid multiple alarms/alerts being forwarded to
the end user(s) of the 'same' fire. A vegetation fire may give rise to more
than one VIIRS pixel detection in space, and if it does not move new detections
at the same location may be triggered at the following overpasses. This
filtering will also avoid repetative alerts of a sustainable fire. We will use
thresholds in space and time to define when a fire is a new fire and when a
pixel-detection is not associated with the same fire. Currently the thresholds
are:

 * Time threshold: 6 hours
 * Spatial threshold: 800 meters
 * Threshold defining the longest possible fire before it will be split into smaller fires: 1.2 km

"""

import os
import logging
import signal
from queue import Empty
from threading import Thread
from requests.exceptions import HTTPError, ConnectionError
from posttroll.listener import ListenerContainer
from posttroll.message import Message
from posttroll.publisher import NoisyPublisher

import pytz
from datetime import datetime, timedelta

from geopy import distance
from geojson import FeatureCollection, dump
from itertools import combinations
from pathlib import Path

from activefires_pp.utils import get_filename_from_posttroll_message
from activefires_pp.config import read_config
from activefires_pp.config import get_xauthentication_token

from activefires_pp.geojson_utils import read_geojson_data
from activefires_pp.geojson_utils import get_geojson_files_in_observation_time_order
from activefires_pp.geojson_utils import store_geojson_alarm
from activefires_pp.api_posting import post_alarm

from trollsift import Parser

LOG = logging.getLogger(__name__)

DIR_SPATIAL_FILTER = '/tmp'


class AlarmFilterRunner(Thread):
    """The Alarm-Filter runner class."""

    def __init__(self, configfile):
        """Initialize the AlarmFilterRunner class."""
        super().__init__()
        self.configfile = configfile
        self.options = {}
        self.time_and_space_thresholds = {}

        self._read_and_check_configuration()

        self.input_topic = self.options['subscribe_topics'][0]
        LOG.debug("Input topic: %s", self.input_topic)
        self.output_topic = self.options['publish_topic']

        self.sos_alarms_file_pattern = self.options['geojson_file_pattern_alarms']
        self.restapi_url = self.options['restapi_url']
        _xauth_filepath = get_xauthentication_filepath_from_environment()
        self._xauth_token = get_xauthentication_token(_xauth_filepath)

        self.fire_alarms_dir = Path(self.options['fire_alarms_dir'])

        self.listener = None
        self.publisher = None
        self.loop = False
        self._setup_and_start_communication()

    def _setup_and_start_communication(self):
        """Set up the Posttroll communication and start the publisher."""
        LOG.debug("Input topic: %s", self.input_topic)
        self.listener = ListenerContainer(topics=[self.input_topic])
        self.publisher = NoisyPublisher("spatiotemporal_alarm_filtering")
        self.publisher.start()
        self.loop = True
        signal.signal(signal.SIGTERM, self.signal_shutdown)

    def _read_and_check_configuration(self):
        """Read and check the configuration parameters."""
        config = read_config(self.configfile)
        self._check_and_set_thresholds_from_config(config)
        self._set_options_from_config(config)

    def _check_and_set_thresholds_from_config(self, config):
        """Check that adequate thresholds are provided in config file."""
        if 'time_and_space_thresholds' not in config:
            raise OSError('Missing time and space thresholds from configuration!')

        self._log_message_per_threshold = {
            'hour_threshold': ("Threshold defining when a new detection should trigger an " +
                               "alarm on the same spot (in hours): ",
                               "Threshold in time is missing!"),
            'long_fires_threshold_km': ("Threshold defining the maximum extention of a fire before " +
                                        "dividing it in smaller pieces (in km): %3.1f",
                                        "Threshold for splitting long fires is missing!"),
            "spatial_threshold_km": ("Threshold defining the maximum distance between two detections " +
                                     "beyond which they must trigger more than one alarm (km): %3.1f",
                                     "A spatial threshold is missing!")
        }
        for thr_type in self._log_message_per_threshold:
            self._check_and_set_threshold(thr_type, config)

    def _check_and_set_threshold(self, thr_type, config):
        """Check if threshold exists in config and set it accordingly or raise an error."""
        if thr_type in config['time_and_space_thresholds']:
            self.time_and_space_thresholds[thr_type] = config['time_and_space_thresholds'][thr_type]
            LOG.info(self._log_message_per_threshold[thr_type][0], self.time_and_space_thresholds[thr_type])
        else:
            raise OSError(self._log_message_per_threshold[thr_type][1])

    def _set_options_from_config(self, config):
        """From the configuration on disk set the option dictionary with all metadata for real-time processing."""
        for item in config:
            self.options[item] = config[item]

        if isinstance(self.options.get('subscribe_topics'), str):
            subscribe_topics = self.options.get('subscribe_topics').split(',')
            for item in subscribe_topics:
                if len(item) == 0:
                    subscribe_topics.remove(item)
            self.options['subscribe_topics'] = subscribe_topics

        if isinstance(self.options.get('publish_topics'), str):
            publish_topics = self.options.get('publish_topics').split(',')
            for item in publish_topics:
                if len(item) == 0:
                    publish_topics.remove(item)
            self.options['publish_topics'] = publish_topics

    def signal_shutdown(self, *args, **kwargs):
        """Shutdown the Notifier process."""
        self.close()

    def run(self):
        """Run the spatiotemporal alarm filtering."""
        while self.loop:
            try:
                msg = self.listener.output_queue.get(timeout=1)
                LOG.debug("Message: %s", str(msg.data))
            except Empty:
                continue
            else:
                if msg.type in ['info', ]:
                    # No fires detected - no notification to send:
                    LOG.info("Message type info: No fires detected - no alarm to generate.")
                    continue
                elif msg.type not in ['file', 'collection', 'dataset']:
                    LOG.debug("Message type not supported: %s", str(msg.type))
                    continue

                generated_alarms = self.spatio_temporal_alarm_filtering(msg)
                if generated_alarms:
                    LOG.debug("Number of generated alarms: %d", len(generated_alarms))
                else:
                    LOG.debug("No alarms generated!")

    def spatio_temporal_alarm_filtering(self, msg):
        """Spatial and temporal filtering of the fire detections and create and post the alarm.

        Perform the spatial and temporal filtering of the current detections
        against a catalogue of past detections and generate an alarm point by
        creating a geoJSON file and posting to a ReST-API.

        """
        LOG.debug("Start the spatio-temporal alarm filtering.")

        filename = get_filename_from_posttroll_message(msg)
        ffdata = read_geojson_data(filename)
        if not ffdata:
            return None

        geojson_alarms = create_alarms_from_fire_detections(ffdata,
                                                            self.fire_alarms_dir, self.sos_alarms_file_pattern,
                                                            self.time_and_space_thresholds)

        if len(geojson_alarms) == 0:
            LOG.info("No alarms to be triggered!")
            return None

        self.send_alarms(geojson_alarms, msg)
        return geojson_alarms

    def send_alarms(self, geojson_alarms, msg):
        """Send the alarms: Create geojson file with alarm data, post it and publish message."""
        p__ = Parser(self.sos_alarms_file_pattern)
        for idx, alarm in enumerate(geojson_alarms):
            # Write alarm to a geojson file in the fire_alarms_dir destination:
            # 1) Create the filename
            # 2) Wite to a file
            output_filename = store_geojson_alarm(self.fire_alarms_dir, p__, idx, alarm)
            try:
                post_alarm(alarm['features'], self.restapi_url, self._xauth_token)
                LOG.info('Alarm sent - status OK')
            except (HTTPError, ConnectionError):
                LOG.exception('Failed sending alarm!')
                LOG.error('Data: %s', str(alarm['features']))

            output_message = _create_output_message(msg, self.output_topic, alarm, output_filename)
            LOG.debug("Sending message: %s", str(output_message))
            self.publisher.send(str(output_message))

    def close(self):
        """Shutdown the AlarmFilterRunner process."""
        LOG.info('Terminating the Alarm Filter Runner process.')
        self.loop = False
        try:
            self.listener.stop()
        except Exception:
            LOG.exception("Couldn't stop listener.")
        if self.publisher:
            try:
                self.publisher.stop()
            except Exception:
                LOG.exception("Couldn't stop publisher.")


def get_xauthentication_filepath_from_environment():
    """Get the filename with the X-Authentication-token from environment."""
    xauth_filepath = os.environ.get('FIREALARMS_XAUTH_FILEPATH')
    if xauth_filepath is None:
        raise OSError("Environment variable FIREALARMS_XAUTH_FILEPATH not set!")

    return xauth_filepath


def dump_collection(idx, features):
    """Dump the list of features as a Geojson Feature Collection."""
    tmpdir = Path(DIR_SPATIAL_FILTER)
    fname = 'sos_alarm_{index}.geojson'.format(index=idx)
    output_filename = tmpdir / fname
    feature_collection = FeatureCollection(features)
    with open(output_filename, 'w') as fpt:
        dump(feature_collection, fpt)


def create_alarms_from_fire_detections(fire_data, past_detections_dir, sos_alarms_file_pattern,
                                       time_space_thresholds):
    """Create alarm(s) from a set of detections."""
    long_fires_threshold = time_space_thresholds.get('long_fires_threshold_km', 1.2)
    gathered_fires = join_fire_detections(fire_data)

    # Now go through the gathered fires and split long/large clusters of
    # detections in smaller parts, and create potential alarms:

    alarms_list = []
    for key in gathered_fires:
        LOG.debug("Key: %s" % key)
        fire_alarms = get_single_point_fires_as_collections(gathered_fires[key], long_fires_threshold)

        alarm_should_be_triggered = False
        for fire_alarm in fire_alarms:
            # Check against the most recent alarms:
            alarm_should_be_triggered = check_if_fire_should_trigger_alarm(fire_alarm, past_detections_dir,
                                                                           sos_alarms_file_pattern,
                                                                           time_space_thresholds)
            if alarm_should_be_triggered:
                lonlat = fire_alarm['features']['geometry']['coordinates']
                power = fire_alarm['features']['properties']['power']
                LOG.info("Alarm should be triggered: Power=%f Location=(%f,%f)", power, lonlat[0], lonlat[1])
                alarms_list.append(fire_alarm)

    return alarms_list


def find_neighbours(feature, other_features, thr_dist=0.8):
    """For a given feature (fire detection) find its close neighbours."""
    lon0, lat0 = feature['geometry']['coordinates']
    idx = []
    for key, feat in other_features.items():
        geom = feat['geometry']
        lon, lat = geom['coordinates']
        km_dist = distance.distance((lat0, lon0), (lat, lon)).kilometers
        if km_dist < thr_dist:
            idx.append(key)

    return idx


def gather_neighbours_to_new_collection(start_id, features, feature_collections, thr_dist=None):
    """Go through all features and gather into groups of neighbouring detections."""
    first_point = features[start_id]
    features.pop(start_id)
    neighbour_ids = find_neighbours(first_point, features)

    lon1, lat1 = first_point['geometry']['coordinates']
    collection_id = "%d_%d" % (lon1*100000, lat1*100000)
    feature_collections[collection_id] = [first_point]
    for i in neighbour_ids:
        if thr_dist:
            lon, lat = features[i]['geometry']['coordinates']
            km_dist = distance.distance((lat1, lon1), (lat, lon)).kilometers
            if km_dist > thr_dist:
                continue

        feature_collections[collection_id].append(features[i])
        features.pop(i)

    return features


def join_fire_detections(gdata):
    """Go through detections and gather detections that are close.

    1) Take the first feature (fire detection)
    2) Identify its neighbours (close fires, with a distance less than some
       threshold, say 800 meters)
    3) Gather those features into the first collection of neighbours
    4) Go through the rest and see if any are close to the collection,
       if there is one, add it to the collection, and repeat.
    5) If there are any features left in the list go to step 1 above, and
       repeat.

    """
    num_features = len(gdata['features'])
    features = dict(zip(range(num_features), gdata['features']))

    feature_collections = {}
    start_index = 0
    while num_features > 0:
        features = gather_neighbours_to_new_collection(start_index, features, feature_collections)
        if len(features.keys()) == 0:
            break
        start_index = min(features.keys())
        num_features = len(features.keys())

    return feature_collections


def split_large_fire_clusters(features, large_fires_threshold):
    """Take a list of fire detection features and split in smaller clusters/chains."""
    num_features = len(features)
    LOG.debug("Split large fire clusters - Number of features: %d" % num_features)
    features = dict(zip(range(num_features), features))

    # Find the two features that are farthest away from each other:
    # First find all 2-combinations of the set:
    two_combs = list(combinations(range(num_features), 2))
    max_distance = 0.0
    max_2comb = (0, 0)
    for tup2 in two_combs:
        lon1, lat1 = features[tup2[0]]['geometry']['coordinates']
        lon2, lat2 = features[tup2[1]]['geometry']['coordinates']

        km_dist = distance.distance((lat1, lon1), (lat2, lon2)).kilometers
        if km_dist > max_distance:
            max_distance = km_dist
            max_2comb = tup2

    if max_distance < large_fires_threshold:
        LOG.debug("Only one cluster - (max_distance, threshold) = (%f, %f)" % (max_distance, large_fires_threshold))
        return {'only-one-cluster': list(features.values())}

    # Now we have located the two detections in the collection that are
    # farthest apart! Then from one of those points gather neighbours to a new
    # collection within threshold distance:
    feature_collections = {}
    start_index = max_2comb[0]
    while num_features > 0:
        features = gather_neighbours_to_new_collection(start_index, features, feature_collections,
                                                       large_fires_threshold)
        if not features:
            break
        start_index = min(features.keys())
        num_features = len(features.keys())

    return feature_collections


def create_one_detection_from_collection(features):
    """From a collection of close fire detections create one detection using the maximum FRP as locator."""
    max_frp = 0.0
    for feat in features:
        if feat["properties"]["power"] > max_frp:
            max_frp = feat["properties"]["power"]
            feature2save = feat

    lonlat = feature2save["geometry"]["coordinates"]
    LOG.info("Fire detection with maxmimum FRP in the cluster found: power = %f location = (%f, %f)",
             feature2save["properties"]["power"], lonlat[0], lonlat[1])
    # Record if there were more than one detection in the collection:
    related_detection = False
    if len(features) > 1:
        related_detection = True

    LOG.info("Related detection: %s" % str(related_detection))
    feature2save["properties"]['related_detection'] = related_detection
    return feature2save


def create_single_point_alarms_from_collections(features):
    """Go through list of features and create a single fire-detection/alarm for each and return as collection."""
    fcollection = []
    for key in features:
        # Make one alarm from each feature-collection (one fire with neighbouring detections):
        single_detection = create_one_detection_from_collection(features[key])

        # Maybe we shouldn't actually store it as a collection now that it is only one detection!?
        fcollection.append(FeatureCollection(single_detection))

    return fcollection


def get_single_point_fires_as_collections(fires, long_fires_threshold):
    """Split larger fires into smaller parts and make a single fire-detection out of each and return as collection."""
    features = split_large_fire_clusters(fires, long_fires_threshold)
    return create_single_point_alarms_from_collections(features)


def check_if_fire_should_trigger_alarm(gjson_data, past_alarms_dir, sos_alarms_file_pattern,
                                       time_space_thresholds):
    """Check if fire point should trigger an alarm.

    The fire point is a GeoJSON object. A search back in time X hours (e.g. X=16) is
    done, and a check for previous fire alarms on that location is done. Only
    if no previous fire alarm at the same position (determined by a distance
    threshold) is found, an alarm should be issued.
    """
    utc = pytz.timezone('utc')
    end_time = datetime.fromisoformat(gjson_data["properties"]["observation_time"])
    end_time = end_time.astimezone(utc).replace(tzinfo=None)

    hour_thr = time_space_thresholds.get('hour_threshold', 16)
    km_threshold = time_space_thresholds.get('spatial_threshold_km', 0.8)

    LOG.debug("Check if fire detection should trigger alarm. Thresholds = (%3.1f h, %3.1f km)",
              hour_thr, km_threshold)

    start_time = end_time - timedelta(hours=hour_thr)
    files_in_observation_time_order = get_geojson_files_in_observation_time_order(past_alarms_dir,
                                                                                  sos_alarms_file_pattern,
                                                                                  (start_time, end_time))

    # If directory is empty there is no history present and this alarm should be triggered:
    if len(files_in_observation_time_order) == 0:
        LOG.info("Directory empty - no history present - alarm should be triggered!")
        return True

    lonlat_current_position = gjson_data["geometry"]["coordinates"]
    shall_trigger_alarm = True
    for filename in reversed(files_in_observation_time_order):
        obstime, dist = distance_and_time_from_geojson_position(lonlat_current_position,
                                                                (past_alarms_dir / filename))
        if dist < km_threshold:
            LOG.debug("Recent alarm file: %s", str(filename))
            shall_trigger_alarm = False

            time_diff_in_minutes = (end_time - obstime).total_seconds()/60.
            LOG.info("There was a recent alarm on this location! No new alarm generated.")
            LOG.info("Distance = %f, Time distance = %f minutes" % (dist, time_diff_in_minutes))
            break

    return shall_trigger_alarm


def distance_and_time_from_geojson_position(position, filepath):
    """Read the geojson data and get the observation time and the distance to a position given as input."""
    lon0, lat0 = position
    gjdata = read_geojson_data(filepath)
    lon, lat = gjdata["geometry"]["coordinates"]
    # Get distance to this fire point:
    dist = distance.distance((lat0, lon0), (lat, lon)).kilometers

    obstime = datetime.fromisoformat(gjdata["properties"]["observation_time"])
    utc = pytz.timezone('utc')
    obstime = obstime.astimezone(utc).replace(tzinfo=None)

    return obstime, dist


def _create_output_message(msg, topic, geojson, filename):
    """Create the output message from the input message and the geojson payload."""
    to_send = msg.data.copy()
    to_send.pop('type', None)
    to_send['related_detection'] = geojson['features']['properties']['related_detection']
    to_send['power'] = geojson['features']['properties']['power']
    to_send['tb'] = geojson['features']['properties']['tb']
    to_send['platform_name'] = geojson['features']['properties']['platform_name']
    to_send['coordinates'] = geojson['features']['geometry']['coordinates']
    to_send['file'] = filename.name
    to_send['format'] = 'geojson'
    to_send['uri'] = str(filename)
    to_send.pop('uid', None)

    return Message(topic, 'file', to_send)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Handling the yaml configurations."""

import yaml
from yaml import UnsafeLoader, Loader


def read_config(config_filepath):
    """Read and extract config information."""
    with open(config_filepath, 'r') as fp_:
        config = yaml.load(fp_, Loader=UnsafeLoader)

    return config


def get_xauthentication_token(xauth_filepath):
    """Get the X-Authentication-token needed for posting to the API."""
    with open(xauth_filepath, 'r') as fp_:
        tokens = yaml.load(fp_, Loader=Loader)

    return tokens['xauth_tokens']['x-auth-satellite-alarm']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c21856.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test getting the yaml configurations from file."""

from activefires_pp.config import read_config
from activefires_pp.config import get_xauthentication_token


def test_get_yaml_configuration(fake_yamlconfig_file):
    """Test read and get the yaml configuration from file."""
    config = read_config(fake_yamlconfig_file)
    assert config['subscribe_topics'] == '/VIIRS/L2/Fires/PP/National'
    assert config['publish_topic'] == '/VIIRS/L2/Fires/PP/SOSAlarm'
    assert config['geojson_file_pattern_alarms'] == 'sos_{start_time:%Y%m%d_%H%M%S}_{id:d}.geojson'
    assert config['fire_alarms_dir'] == '/path/where/the/filtered/alarms/will/be/stored'
    assert config['restapi_url'] == 'https://xxx.smhi.se:xxxx'


def test_get_xauthentication_token(fake_token_file):
    """Test getting the xauthentication token from a file."""
    fake_token = get_xauthentication_token(fake_token_file)
    assert fake_token == 'my-token'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, 2022 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c21856.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the message handling part of the post-processing
"""

from unittest.mock import patch
from datetime import datetime

from posttroll.message import Message
from activefires_pp.post_processing import ActiveFiresPostprocessing
from activefires_pp.spatiotemporal_alarm_filtering import _create_output_message


TEST_MSG = """pytroll://VIIRS/L2/AFI/edr/2/nrk/test/polar/direct_readout file safusr.t@lxserv2313.smhi.se 2021-04-07T00:41:41.568370 v1.01 application/json {"start_time": "2021-04-07T00:28:17", "end_time": "2021-04-07T00:29:40", "orbit_number": 1, "platform_name": "NOAA-20", "sensor": "viirs", "format": "edr", "type": "netcdf", "data_processing_level": "2", "variant": "DR", "orig_orbit_number": 17530, "origin": "172.29.4.164:9099", "uri": "ssh://lxserv2313.smhi.se/san1/polar_out/direct_readout/viirs_active_fires/unfiltered/AFIMG_j01_d20210407_t0028179_e0029407_b17531_c20210407004133375592_cspp_dev.nc", "uid": "AFIMG_j01_d20210407_t0028179_e0029407_b17531_c20210407004133375592_cspp_dev.nc"}"""

CONFIG_EXAMPLE = {'publish_topic': '/VIIRS/L2/Fires/PP',
                  'subscribe_topics': 'VIIRS/L2/AFI',
                  'af_pattern_ibands':
                  'AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S%f}_e{end_hour:%H%M%S%f}_b{orbit:s}_c{processing_time:%Y%m%d%H%M%S%f}_cspp_dev.txt',
                  'geojson_file_pattern_national': 'AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S}.geojson',
                  'geojson_file_pattern_regional': 'AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S}_{region_name:s}.geojson',
                  'regional_shapefiles_format': 'omr_{region_code:s}_Buffer.{ext:s}',
                  'output_dir': '/path/where/the/filtered/results/will/be/stored'}


@patch('socket.gethostname')
@patch('activefires_pp.post_processing.read_config')
@patch('activefires_pp.post_processing.ActiveFiresPostprocessing._setup_and_start_communication')
def test_prepare_posttroll_message(setup_comm, get_config, gethostname):
    """Test setup the posttroll message."""

    get_config.return_value = CONFIG_EXAMPLE
    gethostname.return_value = "my.host.name"

    myconfigfile = "/my/config/file/path"
    myboarders_file = "/my/shape/file/with/country/boarders"
    mymask_file = "/my/shape/file/with/polygons/to/filter/out"

    afpp = ActiveFiresPostprocessing(myconfigfile, myboarders_file, mymask_file)

    test_filepath = "/my/geojson/file/path"

    input_msg = Message.decode(rawstr=TEST_MSG)
    res_msg = afpp._generate_output_message(test_filepath, input_msg)

    assert res_msg.data['platform_name'] == 'NOAA-20'
    assert res_msg.data['type'] == 'GEOJSON-filtered'
    assert res_msg.data['format'] == 'geojson'
    assert res_msg.data['product'] == 'afimg'
    assert res_msg.subject == '/VIIRS/L2/Fires/PP/National'
    assert res_msg.data['uri'] == 'ssh://my.host.name//my/geojson/file/path'

    input_msg = Message.decode(rawstr=TEST_MSG)

    fake_region_mask = {'attributes': {'Kod_omr': '9999',
                                       'Testomr': 'Some area description'}}
    res_msg = afpp._generate_output_message(test_filepath, input_msg, region=fake_region_mask)

    assert res_msg.subject == '/VIIRS/L2/Fires/PP/Regional/9999'

    msg_str = 'No fire detections for this granule'

    result_messages = afpp._generate_no_fires_messages(input_msg, msg_str)

    for (nat_or_reg, res_msg) in zip(['National', 'Regional'], result_messages):
        assert res_msg.data['info'] == msg_str
        assert res_msg.subject == '/VIIRS/L2/Fires/PP/' + nat_or_reg
        assert 'type' not in res_msg.data
        assert 'format' not in res_msg.data
        assert 'product' not in res_msg.data
        assert 'uri' not in res_msg.data


def test_create_output_message(tmp_path):
    """Test create output message from geojson payload."""

    input_msg = Message.decode(rawstr=TEST_MSG)
    filename = tmp_path / 'test_geojson_alarm_file.geojson'
    output_topic = '/VIIRS/L2/Fires/PP/SOSAlarm'
    geojson_alarm = {"features": {"geometry": {"coordinates": [16.249069, 57.156235], "type": "Point"},
                                  "properties": {"confidence": 8, "observation_time": "2021-06-19T02:58:45.700000+02:00",
                                                 "platform_name": "NOAA-20",
                                                 "power": 2.23312426,
                                                 "related_detection": False,
                                                 "tb": 310.37322998}, "type": "Feature"},
                     "type": "FeatureCollection"}

    output_msg = _create_output_message(input_msg, output_topic, geojson_alarm, filename)

    assert output_msg.data == {'start_time': datetime(2021, 4, 7, 0, 28, 17),
                               'end_time': datetime(2021, 4, 7, 0, 29, 40),
                               'orbit_number': 1,
                               'platform_name': 'NOAA-20',
                               'sensor': 'viirs',
                               'data_processing_level': '2',
                               'format': 'geojson',
                               'variant': 'DR',
                               'orig_orbit_number': 17530,
                               'origin': '172.29.4.164:9099',
                               'related_detection': False,
                               'power': 2.23312426,
                               'tb': 310.37322998,
                               'coordinates': [16.249069, 57.156235],
                               'file': 'test_geojson_alarm_file.geojson',
                               'uri': str(filename)}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, 2022 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c21856.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the utility functions.
"""

import pytest
from freezegun import freeze_time
from datetime import datetime, timedelta
from posttroll.message import Message

from activefires_pp.utils import get_filename_from_posttroll_message
from activefires_pp.utils import datetime_utc2local
from activefires_pp.utils import json_serial

NATIONAL_TEST_MESSAGE = """pytroll://VIIRS/L2/Fires/PP/National file safusr.u@lxserv1043.smhi.se 2021-04-19T11:16:49.519087 v1.01 application/json {"start_time": "2021-04-16T12:29:53", "end_time": "2021-04-16T12:31:18", "orbit_number": 1, "platform_name": "NOAA-20", "sensor": "viirs", "data_processing_level": "2", "variant": "DR", "orig_orbit_number": 17666, "uri": "ssh://lxserv1043.smhi.se//san1/polar_out/direct_readout/viirs_active_fires/filtered/AFIMG_j01_d20210416_t122953.geojson", "uid": "AFIMG_j01_d20210416_t122953.geojson", "type": "GEOJSON-filtered", "format": "geojson", "product": "afimg"}"""


def test_json_serial():
    """Test the json_serial function."""

    dtime_obj = datetime(2021, 4, 7, 11, 58, 53, 200000)
    res = json_serial(dtime_obj)

    assert res == "2021-04-07T11:58:53.200000"

    with pytest.raises(TypeError) as exec_info:
        other_obj = 'not okay'
        res = json_serial(other_obj)

    exception_raised = exec_info.value

    assert str(exception_raised) == "Type <class 'str'> not serializable"


def test_get_filename_from_posttroll_message():
    """Test get the filename from the pytroll message data."""
    input_msg = Message.decode(rawstr=NATIONAL_TEST_MESSAGE)
    filename = get_filename_from_posttroll_message(input_msg)

    assert filename.name == 'AFIMG_j01_d20210416_t122953.geojson'
    assert str(filename.parent) == '//san1/polar_out/direct_readout/viirs_active_fires/filtered'


@freeze_time('2022-03-26 18:12:05')
def test_utc2localtime_conversion():
    """Test converting utc time to local time."""

    atime1 = datetime.utcnow()
    dtobj = datetime_utc2local(atime1, 'Europe/Stockholm')
    assert dtobj.strftime('%Y%m%d-%H%M') == '20220326-1912'

    atime2 = atime1 + timedelta(days=1)
    dtobj = datetime_utc2local(atime2, 'Europe/Stockholm')
    assert dtobj.strftime('%Y%m%d-%H%M') == '20220327-2012'

    dtobj = datetime_utc2local(atime1, 'Australia/Sydney')
    assert dtobj.strftime('%Y%m%d-%H%M') == '20220327-0512'

    dtobj2 = datetime_utc2local(atime1, 'Etc/GMT-11')
    assert dtobj2.strftime('%Y%m%d-%H%M') == '20220327-0512'

    dtobj = datetime_utc2local(atime1 + timedelta(days=30), 'Australia/Sydney')
    assert dtobj.strftime('%Y%m%d-%H%M') == '20220426-0412'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Geojson utilities."""

import geojson
import json
import logging
from trollsift import Parser, globify
import pytz
from datetime import datetime
import numpy as np

LOG = logging.getLogger(__name__)


def read_geojson_data(filename):
    """Read Geo json data from file."""
    if str(filename).endswith('.geojson') and filename.exists():
        # Read the file:
        try:
            with open(filename, "r") as fpt:
                return geojson.load(fpt)
        except json.decoder.JSONDecodeError:
            LOG.exception("Geojson file invalid and cannot be read: %s", str(filename))
    else:
        LOG.error("No valid filename to read: %s", str(filename))


def get_geojson_files_in_observation_time_order(path, pattern, time_interval):
    """Get all geojson files with filtered active fire detections (=triggered alarms) since *dtime*."""
    dtime_start = time_interval[0]
    dtime_end = time_interval[1]
    if not dtime_end:
        dtime_end = datetime.utcnow()

    p__ = Parser(pattern)
    files = path.glob(globify(pattern))
    dtimes = []
    fnames = []
    for gjson_file in files:
        fname = gjson_file.name
        res = p__.parse(fname)
        if res['start_time'] > dtime_start and res['start_time'] < dtime_end:
            dtimes.append(res['start_time'])
            fnames.append(fname)

    dtimes = np.array(dtimes)
    fnames = np.array(fnames)

    idx = dtimes.argsort()
    files = np.take(fnames, idx)
    return files.tolist()


def store_geojson_alarm(fires_alarms_dir, file_parser, idx, alarm):
    """Store the fire alarm to a geojson file."""
    utc = pytz.timezone('utc')
    start_time = datetime.fromisoformat(alarm["features"]["properties"]["observation_time"])
    platform_name = alarm["features"]["properties"]["platform_name"]
    start_time = start_time.astimezone(utc).replace(tzinfo=None)
    fname = file_parser.compose({'start_time': start_time, 'id': idx,
                                 'platform_name': platform_name})
    output_filename = fires_alarms_dir / fname
    with open(output_filename, 'w') as fpt:
        geojson.dump(alarm, fpt)

    return output_filename



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021-2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Main module to run the Active Fires Postprocessing in real time.


"""

import logging
import argparse
import sys
from activefires_pp.logger import setup_logging
from activefires_pp.post_processing import ActiveFiresPostprocessing


logger = logging.getLogger('active_fires_postprocessing')


def main():
    """Start and run the active-fires post-processing."""
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--log-config",
                        help="Log config file to use instead of the standard logging.")
    parser.add_argument("-c", "--config",
                        help="YAML config file to use.")
    parser.add_argument("-b", "--shp_boarders",
                        help="Path to shapefile with national boarders",
                        required=True)
    parser.add_argument("-f", "--shp_filtermask",
                        help="Path to shapefile with mask to filter false alarms",
                        required=True)
    parser.add_argument("-r", "--regional_shapefile",
                        help="Path to shapefile with mask to filter detection on regional areas",
                        default=None)
    parser.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0,
                        help="Verbosity (between 1 and 2 occurrences with more leading to more "
                        "verbose logging). WARN=0, INFO=1, "
                        "DEBUG=2. This is overridden by the log config file if specified.")

    cmd_args = parser.parse_args()
    setup_logging(cmd_args)

    configfile = cmd_args.config
    national_map = cmd_args.shp_boarders
    filtermask = cmd_args.shp_filtermask
    regional_filtermask = cmd_args.regional_shapefile

    logger.info("Starting up.")
    try:
        if regional_filtermask:
            fire_pp = ActiveFiresPostprocessing(configfile, national_map, filtermask,
                                                regional_filtermask=regional_filtermask)
        else:
            fire_pp = ActiveFiresPostprocessing(configfile, national_map, filtermask)

    except Exception as err:
        logger.error('Active Fires postprocessing crashed: %s', str(err))
        sys.exit(1)
    try:
        fire_pp.start()
        fire_pp.join()
    except KeyboardInterrupt:
        logger.debug("Interrupting")
    finally:
        fire_pp.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the api-posting part."""


import pytest
import logging
import requests
import responses
import json
from activefires_pp.api_posting import post_alarm

PAST_ALARMS_MONSTERAS3 = """{"features": {"geometry": {"coordinates": [16.252192, 57.15242], "type": "Point"},
"properties": {"confidence": 8, "observation_time": "2021-06-18T14:49:01.750000+02:00",
"platform_name": "NOAA-20", "related_detection": false, "power": 2.87395763, "tb": 330.10293579},
"type": "Feature"}, "type": "FeatureCollection"}"""


@responses.activate
def test_send_alarm_post_ok():
    """Test send alarm."""
    features = json.loads(PAST_ALARMS_MONSTERAS3)
    alarm = features['features']
    restapi_url = "https://my-fake-example.org"

    rsp1 = responses.Response(
        method="POST",
        url=restapi_url,
    )
    responses.add(rsp1)

    post_alarm(alarm, restapi_url)


@responses.activate
def test_send_alarm_post_raise_exception():
    """Test send alarm."""
    features = json.loads(PAST_ALARMS_MONSTERAS3)
    alarm = features['features']
    restapi_url = "https://my-fake-example.org"

    rsp1 = responses.Response(
        method="POST",
        url=restapi_url,
    )
    responses.add(rsp1)
    responses.add(
        responses.POST,
        "http://twitter.com/api/1/foobar",
        status=500,
    )
    with pytest.raises(Exception) as exec_info:
        post_alarm(alarm, "http://twitter.com/api/1/foobar")

    assert exec_info.type == requests.exceptions.HTTPError


@responses.activate
def test_send_alarm_post_log_messages(caplog):
    """Test send alarm."""
    features = json.loads(PAST_ALARMS_MONSTERAS3)
    alarm = features['features']
    restapi_url = "https://my-fake-example.org"

    rsp1 = responses.Response(
        method="POST",
        url=restapi_url,
    )
    responses.add(rsp1)

    with caplog.at_level(logging.INFO):
        _ = post_alarm(alarm, restapi_url)

    log_output = "Alarm posted: Response = <Response [200]>"
    assert log_output in caplog.text



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Set up the activefires-postprocessing package."""

from setuptools import setup
from setuptools import find_packages

try:
    # HACK: https://github.com/pypa/setuptools_scm/issues/190#issuecomment-351181286
    # Stop setuptools_scm from including all repository files
    import setuptools_scm.integration
    setuptools_scm.integration.find_files = lambda _: []
except ImportError:
    pass

with open('./README.md', 'r') as fd:
    long_description = fd.read()

description = 'Post-processing of and notifications on Satellite active fire detections'

requires = ['posttroll', 'netifaces', 'trollsift', 'setuptools_scm', 'pycrs',
            'shapely', 'cartopy', 'pandas', 'geojson', 'fiona', 'geopy', 'matplotlib',
            'requests']
test_requires = ['mock', 'posttroll', 'trollsift', 'pycrs',
                 'shapely', 'cartopy', 'pandas', 'geojson', 'fiona',
                 'freezegun', 'responses']

setup(name="activefires-pp",
      description=description,
      author='Adam Dybroe',
      author_email='adam.dybroe@smhi.se',
      classifiers=["Development Status :: 3 - Alpha",
                   "Intended Audience :: Science/Research",
                   "License :: OSI Approved :: GNU General Public License v3 " +
                   "or later (GPLv3+)",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Topic :: Scientific/Engineering"],
      url="https://github.com/adybbroe/activefires-pp",
      packages=find_packages(),
      long_description=long_description,
      license='GPLv3',
      scripts=['bin/active_fires_postprocessing.py',
               'bin/active_fires_notifier.py',
               'bin/active_fires_spatiotemporal_alarm_filtering.py', ],
      data_files=[],
      zip_safe=False,
      install_requires=requires,
      tests_require=test_requires,
      python_requires='>=3.9',
      use_scm_version=True
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, 2022 Adam.Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the loading of the polygon geometries from the shapefile
"""

import os
from unittest.mock import patch

from activefires_pp.geometries_from_shapefiles import ShapeGeometry

TEST_CRS_PROJ = ('+proj=utm +ellps=GRS80 +a=6378137.0 +rf=298.257222101 +pm=0 +x_0=500000.0 ' +
                 '+y_0=0.0 +lon_0=15.0 +lat_0=0.0 +units=m +axis=enu +no_defs')
TEST_CRS_PROJ_ZONE33 = ('+proj=utm +ellps=GRS80 +a=6378137.0 +rf=298.257222101 +pm=0 +x_0=500000.0 ' +
                        '+y_0=0.0 +lon_0=15.0 +lat_0=0.0 +units=m +axis=enu +no_defs +zone=33')


def fake_shapefiles_glob(dirname):
    list_of_files = []
    for idx in range(4):
        list_of_files.append(os.path.join(dirname, 'some_shape_file_name_%d.shp' % idx))

    return list_of_files


class MyMockProjName(object):
    def __init__(self):
        self.proj4 = 'utm'


class MyMockProj(object):
    def __init__(self):
        self.name = MyMockProjName()


class MyMockCrs(object):
    def __init__(self, crs_name='myname'):
        self._crs_proj = TEST_CRS_PROJ
        self.name = crs_name
        self.proj = MyMockProj()

    def to_proj4(self):
        return self._crs_proj


class fake_geometry_records(object):
    def __init__(self):
        self.geometry = 'Some geom'
        self.attributes = {'attr1': 1, 'attr2': 'myname'}


def fake_get_records():
    """Fake retrieving the generator class of the Geometry records in a shapefile."""
    num = 0
    while num < 10:
        yield fake_geometry_records()
        num = num + 1


@patch('activefires_pp.geometries_from_shapefiles._get_shapefile_paths')
@patch('activefires_pp.geometries_from_shapefiles.pycrs.load.from_file')
def test_shape_geometry_init_single_shapefile_path(load_from_file, get_shapefile_paths):
    """Test creating the ShapeGeometry object from a path to a single shapefile."""

    mypath = '/my/shape/file/path/myshapefile.sph'

    get_shapefile_paths.return_value = [mypath]
    load_from_file.return_value = MyMockCrs('SWEREF99_TM')
    shpgeom = ShapeGeometry(mypath)

    assert shpgeom.proj4str == TEST_CRS_PROJ_ZONE33


@patch('activefires_pp.geometries_from_shapefiles._get_shapefile_paths')
@patch('activefires_pp.geometries_from_shapefiles.pycrs.load.from_file')
def test_shape_geometry_init_dirpath(load_from_file, get_shapefile_paths):
    """Test creating the ShapeGeometry object from a path to a directory with shapefiles."""

    mypath = '/my/shape/file/path/myshapefile.sph'
    get_shapefile_paths.return_value = fake_shapefiles_glob(mypath)

    load_from_file.return_value = MyMockCrs()
    shpgeom = ShapeGeometry(mypath)

    assert shpgeom.proj4str == TEST_CRS_PROJ


@patch('activefires_pp.geometries_from_shapefiles._get_shapefile_paths')
@patch('activefires_pp.geometries_from_shapefiles.pycrs.load.from_file')
def test_shape_geometry_loading(load_from_file, get_shapefile_paths):
    """Test loading the geometries and attributes from the shapefile."""

    mypath = '/my/shape/file/path/myshapefile.sph'
    get_shapefile_paths.return_value = [mypath]

    load_from_file.return_value = MyMockCrs()
    shpgeom = ShapeGeometry(mypath)

    with patch('activefires_pp.geometries_from_shapefiles.shpreader.Reader') as MockShpReader:
        MockShpReader.return_value.records.return_value = fake_get_records()
        shpgeom.load()

    assert shpgeom.geometries is not None
    assert shpgeom.attributes is not None

    assert len(shpgeom.geometries) == 10
    assert len(shpgeom.attributes) == 10



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, 2022 Adam Dybbroe

# Author(s):

#   Adam Dybbroe <Firstname.Lastname at smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the Fires filtering functionality."""

import pytest
from unittest.mock import patch
import pandas as pd
import numpy as np
import io
from datetime import datetime

from activefires_pp.post_processing import ActiveFiresShapefileFiltering
from activefires_pp.post_processing import ActiveFiresPostprocessing
from activefires_pp.post_processing import COL_NAMES


TEST_ACTIVE_FIRES_FILEPATH = "./AFIMG_j01_d20210414_t1126439_e1128084_b17637_c20210414114130392094_cspp_dev.txt"

TEST_ACTIVE_FIRES_FILE_DATA = """
# Active Fires I-band EDR
#
# source: AFIMG_j01_d20210414_t1126439_e1128084_b17637_c20210414114130392094_cspp_dev.nc
# version: CSPP Active Fires version: cspp-active-fire-noaa_1.1.0
#
# column 1: latitude of fire pixel (degrees)
# column 2: longitude of fire pixel (degrees)
# column 3: I04 brightness temperature of fire pixel (K)
# column 4: Along-scan fire pixel resolution (km)
# column 5: Along-track fire pixel resolution (km)
# column 6: detection confidence ([7,8,9]->[lo,med,hi])
# column 7: fire radiative power (MW)
#
# number of fire pixels: 18
#
  59.14783859,   37.85886765,  331.58309937,  0.375,  0.375,    8,    4.67825794
  59.05127335,   28.15227890,  349.83993530,  0.375,  0.375,    8,    7.10289335
  59.05587006,   28.15146446,  326.76165771,  0.375,  0.375,    8,    7.10289335
  59.46587372,   29.04332352,  327.60366821,  0.375,  0.375,    8,    5.01662874
  59.59255981,   28.77226448,  345.88961792,  0.375,  0.375,    8,   13.13724804
  59.58853149,   28.77531433,  339.56134033,  0.375,  0.375,    8,    8.76600266
  59.59326553,   28.77456856,  352.21545410,  0.375,  0.375,    8,    8.76600266
  59.59757233,   28.76391029,  328.43835449,  0.375,  0.375,    8,    5.08633661
  58.35777283,   12.37761784,  327.17175293,  0.375,  0.375,    8,   17.58141518
  60.30867004,   25.53105164,  349.98794556,  0.375,  0.375,    8,    6.93412018
  55.01095581,   -2.28794742,  335.89736938,  0.375,  0.375,    8,    4.39908028
  59.52483368,   17.16816330,  336.57437134,  0.375,  0.375,    8,   14.13167953
  55.00822449,   -2.28098702,  344.50894165,  0.375,  0.375,    8,    4.16644764
  60.13325882,   16.18420029,  329.47689819,  0.375,  0.375,    8,    5.32859230
  61.30901337,   21.98561668,  341.69180298,  0.375,  0.375,    8,    8.87900448
  58.29126740,    0.20132475,  331.47875977,  0.375,  0.375,    8,    3.64687872
  57.42922211,   -3.47403550,  336.02111816,  0.375,  0.375,    8,    8.39092922
  57.42747116,   -3.47912717,  353.80722046,  0.375,  0.375,    8,   12.13035393
"""

CONFIG_EXAMPLE = {'publish_topic': '/VIIRS/L2/Fires/PP',
                  'subscribe_topics': 'VIIRS/L2/AFI',
                  'af_pattern_ibands':
                  'AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S%f}_e{end_hour:%H%M%S%f}' +
                  '_b{orbit:s}_c{processing_time:%Y%m%d%H%M%S%f}_cspp_dev.txt',
                  'geojson_file_pattern_national': 'AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S}.geojson',
                  'geojson_file_pattern_regional': 'AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S}_' +
                  '{region_name:s}.geojson',
                  'regional_shapefiles_format': 'omr_{region_code:s}_Buffer.{ext:s}',
                  'output_dir': '/path/where/the/filtered/results/will/be/stored',
                  'timezone': 'Europe/Stockholm'}

OPEN_FSTREAM = io.StringIO(TEST_ACTIVE_FIRES_FILE_DATA)


MY_FILE_PATTERN = ("AFIMG_{platform:s}_d{start_time:%Y%m%d_t%H%M%S%f}_e{end_hour:%H%M%S%f}_" +
                   "b{orbit:s}_c{processing_time:%Y%m%d%H%M%S%f}_cspp_dev.txt")

TEST_REGIONAL_MASK = {}
TEST_REGIONAL_MASK['Bergslagen (RRB)'] = {'mask': np.array([False, False, False, False, False,
                                                            False, False, False, False,
                                                            False, False, False, False,  True,
                                                            False, False, False, False]),
                                          'attributes': {'Join_Count': 2, 'TARGET_FID': 142,
                                                         'Kod_omr': '1438', 'KNNAMN': 'Dals-Ed',
                                                         'LANDAREAKM': 728.0, 'KNBEF96': 5287.0,
                                                         'OBJECTID': 1080804, 'Datum_Tid': '2016-06-13',
                                                         'Testomr': 'Bergslagen (RRB)',
                                                         'Shape_Leng': 2131994.36042, 'Shape_Area': 53512139344.2},
                                          'all_inside_test_area': False, 'some_inside_test_area': True}
TEST_REGIONAL_MASK['Västerviks kommun'] = {'mask': np.array([False, False, False, False, False,
                                                             False, False, False, False,
                                                             False, False, False, False, False,
                                                             False, False, False, False]),
                                           'attributes': {'Join_Count': 30, 'TARGET_FID': 85,
                                                          'Kod_omr': '0883', 'KNNAMN': 'Västervik',
                                                          'LANDAREAKM': 1870.5, 'KNBEF96': 39579.0,
                                                          'OBJECTID': 1079223, 'Datum_Tid': '2016-06-13',
                                                          'Testomr': 'Västerviks kommun',
                                                          'Shape_Leng': 251653.298274,
                                                          'Shape_Area': 2040770168.02},
                                           'all_inside_test_area': False, 'some_inside_test_area': False}

FAKE_MASK1 = np.array([False, False, False, False, False, False, False, False,  True,
                       False, False,  True, False,  True, False, False, False, False])
FAKE_MASK2 = np.array([True, False,  True])


@patch('activefires_pp.post_processing._read_data')
def test_add_start_and_end_time_to_active_fires_data_utc(readdata):
    """Test adding start and end times to the active fires data."""
    myfilepath = TEST_ACTIVE_FIRES_FILEPATH

    fstream = io.StringIO(TEST_ACTIVE_FIRES_FILE_DATA)
    afdata = pd.read_csv(fstream, index_col=None, header=None, comment='#', names=COL_NAMES)
    readdata.return_value = afdata

    this = ActiveFiresShapefileFiltering(filepath=myfilepath, timezone='GMT')
    with patch('os.path.exists') as mypatch:
        mypatch.return_value = True
        this.get_af_data(filepattern=MY_FILE_PATTERN, localtime=False)

    assert 'starttime' in this._afdata
    assert 'endtime' in this._afdata
    assert this._afdata['starttime'].shape == (18,)

    assert str(this._afdata['starttime'][0]) == '2021-04-14 11:26:43.900000'
    assert str(this._afdata['endtime'][0]) == '2021-04-14 11:28:08'


@patch('activefires_pp.post_processing._read_data')
def test_add_start_and_end_time_to_active_fires_data_localtime(readdata):
    """Test adding start and end times to the active fires data."""
    myfilepath = TEST_ACTIVE_FIRES_FILEPATH

    fstream = io.StringIO(TEST_ACTIVE_FIRES_FILE_DATA)
    afdata = pd.read_csv(fstream, index_col=None, header=None, comment='#', names=COL_NAMES)
    readdata.return_value = afdata

    this = ActiveFiresShapefileFiltering(filepath=myfilepath, timezone='Europe/Stockholm')
    with patch('os.path.exists') as mypatch:
        mypatch.return_value = True
        this.get_af_data(filepattern=MY_FILE_PATTERN, localtime=True)

    assert 'starttime' in this._afdata
    assert 'endtime' in this._afdata

    assert str(this._afdata['starttime'][0]) == '2021-04-14 13:26:43.900000'
    assert str(this._afdata['endtime'][0]) == '2021-04-14 13:28:08'

    this = ActiveFiresShapefileFiltering(filepath=myfilepath, timezone='Iceland')
    with patch('os.path.exists') as mypatch:
        mypatch.return_value = True
        this.get_af_data(filepattern=MY_FILE_PATTERN, localtime=True)

    assert 'starttime' in this._afdata
    assert 'endtime' in this._afdata

    assert str(this._afdata['starttime'][0]) == '2021-04-14 11:26:43.900000'
    assert str(this._afdata['endtime'][0]) == '2021-04-14 11:28:08'

    this = ActiveFiresShapefileFiltering(filepath=myfilepath, timezone='Europe/Helsinki')
    with patch('os.path.exists') as mypatch:
        mypatch.return_value = True
        this.get_af_data(filepattern=MY_FILE_PATTERN, localtime=True)

    assert 'starttime' in this._afdata
    assert 'endtime' in this._afdata

    assert str(this._afdata['starttime'][0]) == '2021-04-14 14:26:43.900000'
    assert str(this._afdata['endtime'][0]) == '2021-04-14 14:28:08'

    this = ActiveFiresShapefileFiltering(filepath=myfilepath, timezone='Europe/Lisbon')
    with patch('os.path.exists') as mypatch:
        mypatch.return_value = True
        this.get_af_data(filepattern=MY_FILE_PATTERN, localtime=True)

    assert 'starttime' in this._afdata
    assert 'endtime' in this._afdata

    assert str(this._afdata['starttime'][0]) == '2021-04-14 12:26:43.900000'
    assert str(this._afdata['endtime'][0]) == '2021-04-14 12:28:08'


@patch('socket.gethostname')
@patch('activefires_pp.post_processing.read_config')
@patch('activefires_pp.post_processing.ActiveFiresPostprocessing._setup_and_start_communication')
def test_regional_fires_filtering(setup_comm, get_config, gethostname):
    """Test the regional fires filtering."""
    # FIXME! This test is to big/broad. Need for refactoring!

    get_config.return_value = CONFIG_EXAMPLE
    gethostname.return_value = "my.host.name"

    myconfigfile = "/my/config/file/path"
    myborders_file = "/my/shape/file/with/country/borders"
    mymask_file = "/my/shape/file/with/polygons/to/filter/out"

    afpp = ActiveFiresPostprocessing(myconfigfile, myborders_file, mymask_file)

    fstream = io.StringIO(TEST_ACTIVE_FIRES_FILE_DATA)
    afdata = pd.read_csv(fstream, index_col=None, header=None, comment='#', names=COL_NAMES)

    # Add metadata to the pandas dataframe:
    fake_metadata = {'platform': 'j01',
                     'start_time': datetime(2021, 4, 14, 11, 26, 43, 900000),
                     'end_hour': datetime(1900, 1, 1, 11, 28, 8, 400000),
                     'orbit': '17637',
                     'processing_time': datetime(2021, 4, 14, 11, 41, 30, 392094),
                     'end_time': datetime(2021, 4, 14, 11, 28, 8)}
    afdata.attrs = fake_metadata

    af_shapeff = ActiveFiresShapefileFiltering(afdata=afdata, platform_name='NOAA-20')
    regional_fmask = TEST_REGIONAL_MASK

    mymsg = "Fake message"
    with patch('activefires_pp.post_processing.store_geojson') as store_geojson:
        with patch('activefires_pp.post_processing.ActiveFiresPostprocessing._generate_output_message') as generate_msg:
            store_geojson.return_value = "/some/output/path"
            generate_msg.return_value = "my fake output message"
            result = afpp.regional_fires_filtering_and_publishing(mymsg, regional_fmask, af_shapeff)

    store_geojson.assert_called_once()
    generate_msg.assert_called_once()

    assert len(result) == 1
    assert result[0] == "my fake output message"


@patch('socket.gethostname')
@patch('activefires_pp.post_processing.read_config')
@patch('activefires_pp.post_processing.ActiveFiresPostprocessing._setup_and_start_communication')
@patch('activefires_pp.post_processing.get_global_mask_from_shapefile', side_effect=[FAKE_MASK1, FAKE_MASK2])
def test_general_national_fires_filtering(get_global_mask, setup_comm, get_config, gethostname):
    """Test the general/basic national fires filtering."""
    get_config.return_value = CONFIG_EXAMPLE
    gethostname.return_value = "my.host.name"

    myconfigfile = "/my/config/file/path"
    myborders_file = "/my/shape/file/with/country/borders"
    mymask_file = "/my/shape/file/with/polygons/to/filter/out"

    afpp = ActiveFiresPostprocessing(myconfigfile, myborders_file, mymask_file)

    fstream = io.StringIO(TEST_ACTIVE_FIRES_FILE_DATA)
    afdata = pd.read_csv(fstream, index_col=None, header=None, comment='#', names=COL_NAMES)

    # Add metadata to the pandas dataframe:
    fake_metadata = {'platform': 'j01',
                     'start_time': datetime(2021, 4, 14, 11, 26, 43, 900000),
                     'end_hour': datetime(1900, 1, 1, 11, 28, 8, 400000),
                     'orbit': '17637',
                     'processing_time': datetime(2021, 4, 14, 11, 41, 30, 392094),
                     'end_time': datetime(2021, 4, 14, 11, 28, 8)}
    afdata.attrs = fake_metadata

    af_shapeff = ActiveFiresShapefileFiltering(afdata=afdata, platform_name='NOAA-20')
    afdata = af_shapeff.get_af_data(MY_FILE_PATTERN)

    mymsg = "Fake message"

    with patch('activefires_pp.post_processing.store_geojson') as store_geojson:
        with patch('activefires_pp.post_processing.ActiveFiresPostprocessing.get_output_messages') as get_output_msg:
            store_geojson.return_value = "/some/output/path"
            get_output_msg.return_value = ["my fake output message"]
            outmsg, result = afpp.fires_filtering(mymsg, af_shapeff)

    store_geojson.assert_called_once()
    get_output_msg.assert_called_once()
    assert get_global_mask.call_count == 2

    assert isinstance(result, pd.core.frame.DataFrame)
    assert len(result) == 1
    np.testing.assert_almost_equal(result.iloc[0]['latitude'], 59.52483368)
    np.testing.assert_almost_equal(result.iloc[0]['longitude'], 17.1681633)
    assert outmsg == ["my fake output message"]


@pytest.mark.usefixtures("fake_national_borders_shapefile")
@pytest.mark.usefixtures("fake_yamlconfig_file_post_processing")
@patch('socket.gethostname')
@patch('activefires_pp.post_processing.ActiveFiresPostprocessing._setup_and_start_communication')
def test_checking_national_borders_shapefile_file_exists(setup_comm, gethostname,
                                                         fake_yamlconfig_file_post_processing,
                                                         fake_national_borders_shapefile):
    """Test the checking of the national borders shapefile - borders shapefile exists."""
    gethostname.return_value = "my.host.name"
    mymask_file = "/my/shape/file/with/polygons/to/filter/out"

    afpp = ActiveFiresPostprocessing(str(fake_yamlconfig_file_post_processing),
                                     fake_national_borders_shapefile, mymask_file)
    afpp._check_borders_shapes_exists()

    assert afpp.shp_borders.name == 'some_national_borders_shape.yaml'
    assert afpp.shp_borders.is_file()


@patch('socket.gethostname')
@patch('activefires_pp.post_processing.read_config')
@patch('activefires_pp.post_processing.ActiveFiresPostprocessing._setup_and_start_communication')
def test_checking_national_borders_shapefile_file_nonexisting(setup_comm, get_config, gethostname):
    """Test the checking of the national borders shapefile - borders shapefile does not exist."""
    get_config.return_value = CONFIG_EXAMPLE
    gethostname.return_value = "my.host.name"

    myconfigfile = "/my/config/file/path"
    myborders_file = "/my/shape/file/with/country/borders"
    mymask_file = "/my/shape/file/with/polygons/to/filter/out"

    afpp = ActiveFiresPostprocessing(myconfigfile, myborders_file, mymask_file)
    with pytest.raises(OSError) as exec_info:
        afpp._check_borders_shapes_exists()

    expected = "Shape file does not exist! Filename = /my/shape/file/with/country/borders"
    assert str(exec_info.value) == expected
